"""Tests for MyBMWVehicle."""
import pytest

from bimmer_connected.const import CarBrands
from bimmer_connected.vehicle import DriveTrainType, VehicleViewDirection

from . import (
    VIN_F11,
    VIN_F31,
    VIN_F35,
    VIN_F44,
    VIN_F45,
    VIN_F48,
    VIN_G01,
    VIN_G05,
    VIN_G08,
    VIN_G21,
    VIN_G30,
    VIN_I01_NOREX,
    VIN_I01_REX,
    get_deprecation_warning_count,
)
from .test_account import account_mock, get_mocked_account

ATTRIBUTE_MAPPING = {
    "remainingFuel": "remaining_fuel",
    "position": "gps_position",
    "cbsData": "condition_based_services",
    "checkControlMessages": "check_control_messages",
    "doorLockState": "door_lock_state",
    "updateReason": "last_update_reason",
    "chargingLevelHv": "charging_level_hv",
    "chargingStatus": "charging_status",
    "maxRangeElectric": "max_range_electric",
    "remainingRangeElectric": "remaining_range_electric",
    "parkingLight": "parking_lights",
    "remainingRangeFuel": "remaining_range_fuel",
    "updateTime": "timestamp",
    "chargingTimeRemaining": "charging_time_remaining",
}


@pytest.mark.asyncio
async def test_drive_train(caplog):
    """Tests around drive_train attribute."""
    vehicle = (await get_mocked_account()).get_vehicle(VIN_G21)
    assert DriveTrainType.PLUGIN_HYBRID == vehicle.drive_train

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_parsing_attributes(caplog):
    """Test parsing different attributes of the vehicle."""
    account = await get_mocked_account()

    for vehicle in account.vehicles:
        print(vehicle.name)
        assert vehicle.drive_train is not None
        assert vehicle.name is not None
        assert isinstance(vehicle.brand, CarBrands)
        assert vehicle.has_combustion_drivetrain is not None
        assert vehicle.has_electric_drivetrain is not None
        assert vehicle.drive_train_attributes is not None
        assert vehicle.is_charging_plan_supported is not None

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_drive_train_attributes(caplog):
    """Test parsing different attributes of the vehicle."""
    account = await get_mocked_account()

    vehicle_drivetrains = {
        VIN_F11: (True, False, False),
        VIN_F31: (True, False, False),
        VIN_F35: (True, False, False),
        VIN_F44: (True, False, False),
        VIN_F45: (True, True, False),
        VIN_F48: (True, False, False),
        VIN_G01: (True, True, False),
        VIN_G05: (True, True, False),
        VIN_G08: (False, True, False),
        VIN_G21: (True, True, False),
        VIN_G30: (True, True, False),
        VIN_I01_NOREX: (False, True, False),
        VIN_I01_REX: (False, True, True),
    }

    for vehicle in account.vehicles:
        assert vehicle_drivetrains[vehicle.vin][0] == vehicle.has_combustion_drivetrain
        assert vehicle_drivetrains[vehicle.vin][1] == vehicle.has_electric_drivetrain
        assert vehicle_drivetrains[vehicle.vin][2] == vehicle.has_range_extender_drivetrain

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_parsing_of_lsc_type(caplog):
    """Test parsing the lsc type field."""
    account = await get_mocked_account()

    for vehicle in account.vehicles:
        assert vehicle.lsc_type is not None

    assert len(get_deprecation_warning_count(caplog)) == 0


def test_car_brand(caplog):
    """Test CarBrand enum"""
    assert CarBrands("BMW") == CarBrands("bmw")

    with pytest.raises(ValueError):
        CarBrands("Audi")

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_get_is_tracking_enabled(caplog):
    """Test setting observer position"""
    vehicle = (await get_mocked_account()).get_vehicle(VIN_F11)
    assert vehicle.is_vehicle_tracking_enabled is False

    vehicle = (await get_mocked_account()).get_vehicle(VIN_F31)
    assert vehicle.is_vehicle_tracking_enabled is True

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_available_attributes(caplog):
    """Check that available_attributes returns exactly the arguments we have in our test data."""
    account = await get_mocked_account()

    vehicle = account.get_vehicle(VIN_F31)
    assert ["gps_position", "vin"] == vehicle.available_attributes

    vehicle = account.get_vehicle(VIN_G08)
    assert [
        "gps_position",
        "vin",
        "remaining_range_total",
        "mileage",
        "charging_time_remaining",
        "charging_start_time",
        "charging_end_time",
        "charging_time_label",
        "charging_status",
        "charging_level_hv",
        "connection_status",
        "remaining_range_electric",
        "last_charging_end_result",
        "condition_based_services",
        "check_control_messages",
        "door_lock_state",
        "timestamp",
        "last_update_reason",
        "lids",
        "windows",
    ] == vehicle.available_attributes

    vehicle = account.get_vehicle(VIN_G30)
    assert [
        "gps_position",
        "vin",
        "remaining_range_total",
        "mileage",
        "charging_time_remaining",
        "charging_start_time",
        "charging_end_time",
        "charging_time_label",
        "charging_status",
        "charging_level_hv",
        "connection_status",
        "remaining_range_electric",
        "last_charging_end_result",
        "remaining_fuel",
        "remaining_range_fuel",
        "fuel_percent",
        "condition_based_services",
        "check_control_messages",
        "door_lock_state",
        "timestamp",
        "last_update_reason",
        "lids",
        "windows",
    ] == vehicle.available_attributes

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_vehicle_image(caplog):
    """Test vehicle image request."""
    vehicle = (await get_mocked_account()).get_vehicle(VIN_G05)

    with account_mock() as mock_api:
        mock_api.get(
            path__regex=r"(.*)/eadrax-ics/v3/presentation/vehicles/\w*/images",
            params={"carView": "VehicleInfo"},
            headers={"accept": "image/png"},
        ).respond(200, content="png_image")
        assert b"png_image" == await vehicle.get_vehicle_image(VehicleViewDirection.FRONT)

    assert len(get_deprecation_warning_count(caplog)) == 0



"""Trigger remote services on a vehicle."""

import asyncio
import datetime
import json
import logging
from typing import TYPE_CHECKING, Dict, Union

from bimmer_connected.api.client import MyBMWClient
from bimmer_connected.const import (
    REMOTE_SERVICE_POSITION_URL,
    REMOTE_SERVICE_STATUS_URL,
    REMOTE_SERVICE_URL,
    VEHICLE_POI_URL,
)
from bimmer_connected.utils import MyBMWJSONEncoder
from bimmer_connected.vehicle.models import PointOfInterest, StrEnum

if TYPE_CHECKING:
    from bimmer_connected.vehicle import MyBMWVehicle

TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"

_LOGGER = logging.getLogger(__name__)

#: time in seconds between polling updates on the status of a remote service
_POLLING_CYCLE = 3

#: maximum number of seconds to wait for the server to return a positive answer
_POLLING_TIMEOUT = 240


class ExecutionState(StrEnum):
    """Enumeration of possible states of the execution of a remote service."""

    INITIATED = "INITIATED"
    PENDING = "PENDING"
    DELIVERED = "DELIVERED"
    EXECUTED = "EXECUTED"
    ERROR = "ERROR"
    UNKNOWN = "UNKNOWN"


class Services(StrEnum):
    """Enumeration of possible services to be executed."""

    LIGHT_FLASH = "light-flash"
    VEHICLE_FINDER = "vehicle-finder"
    DOOR_LOCK = "door-lock"
    DOOR_UNLOCK = "door-unlock"
    HORN = "horn-blow"
    AIR_CONDITIONING = "climate-now"
    CHARGE_NOW = "CHARGE_NOW"


class RemoteServiceStatus:  # pylint: disable=too-few-public-methods
    """Wraps the status of the execution of a remote service."""

    def __init__(self, response: dict):
        """Construct a new object from a dict."""
        status = None
        if "eventStatus" in response:
            status = response.get("eventStatus")

        self.state = ExecutionState(status or "UNKNOWN")
        self.details = response


class RemoteServices:
    """Trigger remote services on a vehicle."""

    def __init__(self, vehicle: "MyBMWVehicle"):
        """Constructor."""
        self._account = vehicle.account
        self._vehicle = vehicle

    async def trigger_remote_light_flash(self) -> RemoteServiceStatus:
        """Trigger the vehicle to flash its headlights.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug("Triggering remote light flash")
        return await self.trigger_remote_service(Services.LIGHT_FLASH)

    async def trigger_remote_door_lock(self) -> RemoteServiceStatus:
        """Trigger the vehicle to lock its doors.

        A state update is triggered after this, as the lock state of the vehicle changes.
        """
        _LOGGER.debug("Triggering remote door lock")
        result = await self.trigger_remote_service(Services.DOOR_LOCK)
        await self._trigger_state_update()
        return result

    async def trigger_remote_door_unlock(self) -> RemoteServiceStatus:
        """Trigger the vehicle to unlock its doors.

        A state update is triggered after this, as the lock state of the vehicle changes.
        """
        _LOGGER.debug("Triggering remote door unlock")
        result = await self.trigger_remote_service(Services.DOOR_UNLOCK)
        await self._trigger_state_update()
        return result

    async def trigger_remote_horn(self) -> RemoteServiceStatus:
        """Trigger the vehicle to sound its horn.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug("Triggering remote horn sound")
        return await self.trigger_remote_service(Services.HORN)

    async def trigger_charge_now(self) -> RemoteServiceStatus:
        """Trigger the vehicle to start charging.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug("Triggering charge now")
        result = await self.trigger_remote_service(Services.CHARGE_NOW)
        await self._trigger_state_update()
        return result

    async def trigger_remote_air_conditioning(self) -> RemoteServiceStatus:
        """Trigger the air conditioning to start.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug("Triggering remote air conditioning")
        result = await self.trigger_remote_service(Services.AIR_CONDITIONING, {"action": "START"})
        await self._trigger_state_update()
        return result

    async def trigger_remote_air_conditioning_stop(self) -> RemoteServiceStatus:
        """Trigger the air conditioning to stop.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug("Triggering remote air conditioning")
        result = await self.trigger_remote_service(Services.AIR_CONDITIONING, {"action": "STOP"})
        await self._trigger_state_update()
        return result

    async def trigger_remote_service(self, service_id: Services, params: Dict = None) -> RemoteServiceStatus:
        """Trigger a generic remote service and wait for the result."""
        event_id = await self._start_remote_service(service_id, params)
        status = await self._block_until_done(event_id)
        return status

    async def _start_remote_service(self, service_id: Services, params: Dict = None) -> str:
        """Start a generic remote service."""

        url = REMOTE_SERVICE_URL.format(vin=self._vehicle.vin, service_type=service_id.value)
        async with MyBMWClient(self._account.mybmw_client_config, brand=self._vehicle.brand) as client:
            response = await client.post(url, params=params)
        return response.json().get("eventId")

    async def _block_until_done(self, event_id: str) -> RemoteServiceStatus:
        """Keep polling the server until we get a final answer.

        :raises TimeoutError: if there is no final answer before _POLLING_TIMEOUT
        """
        fail_after = datetime.datetime.now() + datetime.timedelta(seconds=_POLLING_TIMEOUT)
        while datetime.datetime.now() < fail_after:
            await asyncio.sleep(_POLLING_CYCLE)
            status = await self._get_remote_service_status(event_id)
            _LOGGER.debug("current state of '%s' is: %s", event_id, status.state.value)
            if status.state == ExecutionState.ERROR:
                raise Exception(f"Remote service failed with state '{status.state}'. Response: {status.details}")
            if status.state not in [ExecutionState.UNKNOWN, ExecutionState.PENDING, ExecutionState.DELIVERED]:
                return status
        raise TimeoutError(
            f"Did not receive remote service result for '{event_id}' in {_POLLING_TIMEOUT} seconds. "
            f"Current state: {status.state.value}"
        )

    async def _get_remote_service_status(self, event_id: str = None) -> RemoteServiceStatus:
        """The execution status of the last remote service that was triggered."""
        _LOGGER.debug("getting remote service status for '%s'", event_id)
        url = REMOTE_SERVICE_STATUS_URL.format(vin=self._vehicle.vin, event_id=event_id)
        async with MyBMWClient(self._account.mybmw_client_config, brand=self._vehicle.brand) as client:
            response = await client.post(url)
        return RemoteServiceStatus(response.json())

    async def _trigger_state_update(self) -> None:
        """Sleep for 2x POLLING_CYCLE and force-refresh vehicles from BMW servers."""
        await asyncio.sleep(_POLLING_CYCLE * 2)
        await self._account.get_vehicles()

    async def trigger_send_poi(self, poi: Union[PointOfInterest, Dict]) -> RemoteServiceStatus:
        """Send a PointOfInterest to the vehicle.

        :param poi: A PointOfInterest containing at least 'lat' and 'lon' and optionally
            'name', 'street', 'city', 'postalCode', 'country'

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug("Sending PointOfInterest to car")

        if isinstance(poi, Dict):
            poi = PointOfInterest(**poi)

        async with MyBMWClient(self._account.mybmw_client_config, brand=self._vehicle.brand) as client:
            await client.post(
                VEHICLE_POI_URL,
                headers={"content-type": "application/json"},
                data=json.dumps(
                    {
                        "location": poi.__dict__,
                        "vin": self._vehicle.vin,
                    },
                    cls=MyBMWJSONEncoder,
                ),
            )

        # send-to-car has no separate ExecutionStates
        return RemoteServiceStatus({"eventStatus": "EXECUTED"})

    async def trigger_remote_vehicle_finder(self) -> RemoteServiceStatus:
        """Trigger the vehicle finder.

        A state update is triggered after this, as the location state of the vehicle changes.
        """
        _LOGGER.debug("Triggering remote vehicle finder")
        event_id = await self._start_remote_service(Services.VEHICLE_FINDER)
        status = await self._block_until_done(event_id)
        result = await self._get_event_position(event_id)
        self._vehicle.vehicle_location.set_remote_service_position(result)
        return status

    async def _get_event_position(self, event_id) -> Dict:
        url = REMOTE_SERVICE_POSITION_URL.format(event_id=event_id)
        if not self._account.observer_position:
            return {
                "errorDetails": {
                    "title": "Unknown position",
                    "description": "Set observer position to retrieve vehicle coordinates!",
                }
            }
        async with MyBMWClient(self._account.mybmw_client_config, brand=self._vehicle.brand) as client:
            response = await client.post(
                url,
                headers={
                    "latitude": str(self._account.observer_position.latitude),
                    "longitude": str(self._account.observer_position.longitude),
                },
            )
        return response.json()



"""Tests for utils."""
import datetime
import json
import os
import time
from unittest import mock

import pytest
import time_machine

from bimmer_connected.utils import MyBMWJSONEncoder, get_class_property_names, parse_datetime

from . import RESPONSE_DIR, VIN_G21, get_deprecation_warning_count
from .test_account import get_mocked_account


@pytest.mark.asyncio
async def test_drive_train():
    """Tests available attribute."""
    vehicle = (await get_mocked_account()).get_vehicle(VIN_G21)
    assert [
        "available_attributes",
        "brand",
        "drive_train",
        "drive_train_attributes",
        "fuel_indicator_count",
        "has_combustion_drivetrain",
        "has_electric_drivetrain",
        "has_hv_battery",
        "has_internal_combustion_engine",
        "has_range_extender",
        "has_range_extender_drivetrain",
        "has_weekly_planner_service",
        "is_charging_plan_supported",
        "is_lsc_enabled",
        "is_vehicle_active",
        "is_vehicle_tracking_enabled",
        "last_update_reason",
        "lsc_type",
        "mileage",
        "name",
        "timestamp",
        "vin",
    ] == get_class_property_names(vehicle)


@time_machine.travel("2011-11-28 21:28:59 +0000", tick=False)
@pytest.mark.asyncio
async def test_to_json(caplog):
    """Test serialization to JSON."""
    with mock.patch("bimmer_connected.account.MyBMWAccount.timezone", new_callable=mock.PropertyMock) as mock_timezone:
        # Force UTC
        os.environ["TZ"] = "UTC"
        time.tzset()
        mock_timezone.return_value = datetime.timezone.utc

        account = await get_mocked_account()
        vehicle = account.get_vehicle(VIN_G21)

        # Unset UTC after vehicle has been loaded
        del os.environ["TZ"]
        time.tzset()

        with open(RESPONSE_DIR / "G21" / "json_export.json", "rb") as file:
            expected = file.read().decode("UTF-8")

        expected_lines = expected.split("\n")
        actual_lines = json.dumps(vehicle, cls=MyBMWJSONEncoder, indent=4).split("\n")

        for i in range(max(len(expected_lines), len(actual_lines))):
            assert expected_lines[i] == actual_lines[i], f"line {i+1}"

        assert len(get_deprecation_warning_count(caplog)) == 0


def test_parse_datetime(caplog):
    """Test datetime parser."""

    dt_with_milliseconds = datetime.datetime(2021, 11, 12, 13, 14, 15, 567000, tzinfo=datetime.timezone.utc)
    dt_without_milliseconds = datetime.datetime(2021, 11, 12, 13, 14, 15, tzinfo=datetime.timezone.utc)

    assert dt_with_milliseconds == parse_datetime("2021-11-12T13:14:15.567Z")

    assert dt_without_milliseconds == parse_datetime("2021-11-12T13:14:15Z")

    unparseable_datetime = "2021-14-12T13:14:15Z"
    assert parse_datetime(unparseable_datetime) is None
    errors = [r for r in caplog.records if r.levelname == "ERROR" and unparseable_datetime in r.message]
    assert len(errors) == 1



"""Generals models used for bimmer_connected."""


import datetime
import logging
from dataclasses import dataclass
from typing import Any, Dict, Optional

from bimmer_connected.vehicle.models import StrEnum, ValueWithUnit, VehicleDataBase

_LOGGER = logging.getLogger(__name__)


class ChargingState(StrEnum):
    """Charging state of electric vehicle."""

    DEFAULT = "DEFAULT"
    CHARGING = "CHARGING"
    ERROR = "ERROR"
    COMPLETE = "COMPLETE"
    FULLY_CHARGED = "FULLY_CHARGED"
    FINISHED_FULLY_CHARGED = "FINISHED_FULLY_CHARGED"
    FINISHED_NOT_FULL = "FINISHED_NOT_FULL"
    INVALID = "INVALID"
    NOT_CHARGING = "NOT_CHARGING"
    PLUGGED_IN = "PLUGGED_IN"
    WAITING_FOR_CHARGING = "WAITING_FOR_CHARGING"


@dataclass
class FuelAndBattery(VehicleDataBase):  # pylint:disable=too-many-instance-attributes
    """Provides an accessible version of `status.FuelAndBattery`."""

    remaining_range_fuel: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Get the remaining range of the vehicle on fuel."""

    remaining_range_electric: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Get the remaining range of the vehicle on electricity."""

    remaining_range_total: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Get the total remaining range of the vehicle (fuel + electricity, if available)."""

    remaining_fuel: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Get the remaining fuel of the vehicle."""

    remaining_fuel_percent: Optional[int] = None
    """State of charge of the high voltage battery in percent."""

    remaining_battery_percent: Optional[int] = None
    """State of charge of the high voltage battery in percent."""

    charging_status: Optional[ChargingState] = None
    """Charging state of the vehicle."""

    charging_start_time: Optional[datetime.datetime] = None
    """The planned time the vehicle will start charging."""

    charging_end_time: Optional[datetime.datetime] = None
    """The estimated time the vehicle will have finished charging."""

    charging_time_label: Optional[str] = None
    """The planned start/estimated end time as provided by the API."""

    is_charger_connected: bool = False
    """Get status of the connection"""

    account_timezone: datetime.timezone = datetime.timezone.utc

    # pylint:disable=arguments-differ
    @classmethod
    def from_vehicle_data(cls, vehicle_data: Dict):
        """Creates the class based on vehicle data from API."""
        parsed = cls._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return cls(**parsed)
        return None

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Optional[Dict]:
        """Parse fuel indicators based on Ids."""
        retval: Dict[str, Any] = {}

        properties = vehicle_data.get("properties", {})

        fuel_level = properties.get("fuelLevel", {})
        retval["remaining_fuel"] = ValueWithUnit(
            fuel_level.get("value"),
            fuel_level.get("units"),
        )

        if properties.get("fuelPercentage", {}).get("value"):
            retval["remaining_fuel_percent"] = int(properties.get("fuelPercentage", {}).get("value"))

        if properties.get("electricRangeAndStatus", {}).get("chargePercentage"):
            retval["remaining_battery_percent"] = int(
                properties.get("electricRangeAndStatus", {}).get("chargePercentage")
            )

        retval["is_charger_connected"] = properties.get("chargingState", {}).get("isChargerConnected", False)

        fuel_indicators = vehicle_data.get("status", {}).get("fuelIndicators", [])
        for indicator in fuel_indicators:
            if (indicator.get("rangeIconId") or indicator.get("infoIconId")) == 59691:  # Combined
                retval["remaining_range_total"] = cls._parse_to_tuple(indicator)
            elif (indicator.get("rangeIconId") or indicator.get("infoIconId")) == 59683:  # Electric
                retval["remaining_range_electric"] = cls._parse_to_tuple(indicator)

                retval["charging_time_label"] = indicator["infoLabel"]
                retval["charging_status"] = ChargingState(indicator["chargingStatusType"])

                if indicator.get("chargingStatusType") in ["CHARGING", "PLUGGED_IN"]:
                    retval.update(cls._parse_charging_timestamp(indicator))

            elif (indicator.get("rangeIconId") or indicator.get("infoIconId")) == 59681:  # Fuel
                retval["remaining_range_fuel"] = cls._parse_to_tuple(indicator)

        retval["remaining_range_total"] = (
            retval.get("remaining_range_total")
            or retval.get("remaining_range_fuel")
            or retval.get("remaining_range_electric")
        )

        return retval

    def _update_after_parse(self, parsed: Dict) -> Dict:
        """Updates parsed vehicle data with attributes stored in class if needed."""
        if parsed.get("charging_end_time"):
            parsed["charging_end_time"] = parsed["charging_end_time"].replace(tzinfo=self.account_timezone)
        if parsed.get("charging_start_time"):
            parsed["charging_start_time"] = parsed["charging_start_time"].replace(tzinfo=self.account_timezone)
        return parsed

    @staticmethod
    def _parse_charging_timestamp(indicator: Dict) -> Dict:
        """Parse charging end time string to timestamp."""
        charging_start_time: Optional[datetime.datetime] = None
        charging_end_time: Optional[datetime.datetime] = None

        # Only calculate charging end time if infolabel is like '100% at ~11:04am'
        # Other options: 'Charging', 'Starts at ~09:00am' (but not handled here)

        time_str = indicator["infoLabel"].split("~")[-1].strip()
        try:
            time_parsed = datetime.datetime.strptime(time_str, "%I:%M %p")

            current_time = datetime.datetime.now()
            datetime_parsed = time_parsed.replace(
                year=current_time.year, month=current_time.month, day=current_time.day
            )
            if datetime_parsed < current_time:
                datetime_parsed = datetime_parsed + datetime.timedelta(days=1)

            if indicator["chargingStatusType"] == "CHARGING":
                charging_end_time = datetime_parsed
            elif indicator["chargingStatusType"] == "PLUGGED_IN":
                charging_start_time = datetime_parsed
        except ValueError:
            _LOGGER.error("Error parsing charging end time '%s' out of '%s'", time_str, indicator["infoLabel"])
        return {
            "charging_end_time": charging_end_time,
            "charging_start_time": charging_start_time,
        }

    @staticmethod
    def _parse_to_tuple(fuel_indicator):
        """Parse fuel indicator to standard range tuple."""
        try:
            range_val = int(fuel_indicator["rangeValue"])
        except ValueError:
            return ValueWithUnit(None, None)
        return ValueWithUnit(range_val, fuel_indicator["rangeUnits"])



"""Utils for bimmer_connected.api."""

import base64
import hashlib
import json
import logging
import pathlib
import random
import string
from typing import Dict, List, Union

import httpx

UNICODE_CHARACTER_SET = string.ascii_letters + string.digits + "-._~"


def generate_token(length: int = 30, chars: str = UNICODE_CHARACTER_SET) -> str:
    """Generate a random token with given length and characters."""
    rand = random.SystemRandom()
    return "".join(rand.choice(chars) for _ in range(length))


def create_s256_code_challenge(code_verifier: str) -> str:
    """Create S256 code_challenge with the given code_verifier."""
    data = hashlib.sha256(code_verifier.encode("ascii")).digest()
    return base64.urlsafe_b64encode(data).rstrip(b"=").decode("UTF-8")


async def raise_for_status_event_handler(response: httpx.Response):
    """Event handler that automatically raises HTTPStatusErrors when attached.

    Will only raise on 4xx/5xx errors and not raise on 3xx.
    """
    if response.is_error:
        await response.aread()
        response.raise_for_status()


def handle_http_status_error(
    ex: httpx.HTTPStatusError, module: str = "MyBMW API", log_handler: logging.Logger = None
) -> None:
    """Try to extract information from response and re-raise Exception."""
    _logger = log_handler or logging.getLogger(__name__)
    try:
        err = ex.response.json()
        _logger.error("%s error (%s): %s", module, err["error"], err["error_description"])
    except (json.JSONDecodeError, KeyError):
        _logger.error("%s error: %s", module, ex.response.text)
    raise ex


def anonymize_data(json_data: Union[List, Dict]) -> Union[List, Dict]:
    """Replace parts of the logfiles containing personal information."""

    replacements = {
        "lat": 12.3456,
        "latitude": 12.3456,
        "lon": 34.5678,
        "longitude": 34.5678,
        "heading": 123,
        "vin": "some_vin",
        "licensePlate": "some_license_plate",
        "name": "some_name",
        "city": "some_city",
        "street": "some_street",
        "streetNumber": "999",
        "postalCode": "some_postal_code",
        "phone": "some_phone",
        "formatted": "some_formatted_address",
        "subtitle": "some_road \u2022 duration \u2022 -- EUR",
    }

    if isinstance(json_data, list):
        json_data = [anonymize_data(v) for v in json_data]
    elif isinstance(json_data, dict):
        for key, value in json_data.items():
            if key in replacements:
                json_data[key] = replacements[key]
            else:
                json_data[key] = anonymize_data(value)

    return json_data


def log_to_to_file(content: Union[str, bytes, List, Dict], logfile_path: pathlib.Path, logfile_name: str) -> None:
    """If a log path is set, log all responses to a file."""
    if logfile_path is None or logfile_name is None:
        return

    try:
        parsed_content = json.loads(content)  # type: ignore[arg-type]
        anonymized_data = json.dumps(anonymize_data(parsed_content), indent=2, sort_keys=True)  # type: ignore[arg-type]
        file_extension = "json"
    except json.JSONDecodeError:
        anonymized_data = content.decode("UTF-8") if isinstance(content, bytes) else content  # type: ignore[assignment]
        file_extension = "txt"

    output_path = None
    count = 0

    while output_path is None or output_path.exists():
        output_path = logfile_path / f"{logfile_name}_{count}.{file_extension}"
        count += 1

    with open(output_path, "w", encoding="UTF-8") as logfile:
        logfile.write(anonymized_data)



"""Test for remote_services."""
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import Dict, List

try:
    from unittest import mock

    if not hasattr(mock, "AsyncMock"):
        # AsyncMock was only introduced with Python3.8, so we have to use the backported module
        raise ImportError()
except ImportError:
    import mock  # type: ignore[import,no-redef]

from uuid import uuid4

import httpx
import pytest
import time_machine

from bimmer_connected.vehicle import remote_services
from bimmer_connected.vehicle.remote_services import ExecutionState, RemoteServiceStatus

from . import RESPONSE_DIR, VIN_F45, load_response
from .test_account import account_mock, get_mocked_account

_RESPONSE_INITIATED = RESPONSE_DIR / "remote_services" / "eadrax_service_initiated.json"
_RESPONSE_PENDING = RESPONSE_DIR / "remote_services" / "eadrax_service_pending.json"
_RESPONSE_DELIVERED = RESPONSE_DIR / "remote_services" / "eadrax_service_delivered.json"
_RESPONSE_EXECUTED = RESPONSE_DIR / "remote_services" / "eadrax_service_executed.json"
_RESPONSE_ERROR = RESPONSE_DIR / "remote_services" / "eadrax_service_error.json"
_RESPONSE_EVENTPOSITION = RESPONSE_DIR / "remote_services" / "eadrax_service_eventposition.json"


POI_DATA = {
    "lat": 37.4028943,
    "lon": -121.9700289,
    "name": "49ers",
    "street": "4949 Marie P DeBartolo Way",
    "city": "Santa Clara",
    "postal_code": "CA 95054",
    "country": "United States",
}

STATUS_RESPONSE_ORDER = [_RESPONSE_PENDING, _RESPONSE_DELIVERED, _RESPONSE_EXECUTED]
STATUS_RESPONSE_DICT: Dict[str, List[Path]] = defaultdict(lambda: deepcopy(STATUS_RESPONSE_ORDER))


def service_trigger_sideeffect(request: httpx.Request) -> httpx.Response:  # pylint: disable=unused-argument
    """Returns specific eventId for each remote function."""
    json_data = load_response(_RESPONSE_INITIATED)
    json_data["eventId"] = str(uuid4())
    return httpx.Response(200, json=json_data)


def service_status_sideeffect(request: httpx.Request) -> httpx.Response:
    """Returns all 3 eventStatus responses per function."""
    response_data = STATUS_RESPONSE_DICT[request.url.params["eventId"]].pop(0)
    return httpx.Response(200, json=load_response(response_data))


def remote_services_mock():
    """Returns mocked adapter for auth."""
    router = account_mock()

    router.post(path__regex=r"/eadrax-vrccs/v2/presentation/remote-commands/.+/.+$").mock(
        side_effect=service_trigger_sideeffect
    )
    router.post("/eadrax-vrccs/v2/presentation/remote-commands/eventStatus", params={"eventId": mock.ANY}).mock(
        side_effect=service_status_sideeffect
    )

    router.post("/eadrax-dcs/v1/send-to-car/send-to-car").respond(201)
    router.post("/eadrax-vrccs/v2/presentation/remote-commands/eventPosition", params={"eventId": mock.ANY}).respond(
        200,
        json=load_response(_RESPONSE_EVENTPOSITION),
    )
    return router


def test_states():
    """Test parsing the different response types."""
    rss = RemoteServiceStatus(load_response(_RESPONSE_PENDING))
    assert ExecutionState.PENDING == rss.state

    rss = RemoteServiceStatus(load_response(_RESPONSE_DELIVERED))
    assert ExecutionState.DELIVERED == rss.state

    rss = RemoteServiceStatus(load_response(_RESPONSE_EXECUTED))
    assert ExecutionState.EXECUTED == rss.state


@remote_services_mock()
@pytest.mark.asyncio
@pytest.mark.filterwarnings("ignore:coroutine 'AsyncMockMixin._execute_mock_call' was never awaited:RuntimeWarning")
async def test_trigger_remote_services():
    """Test executing a remote light flash."""
    remote_services._POLLING_CYCLE = 0  # pylint: disable=protected-access

    services = [
        ("LIGHT_FLASH", "trigger_remote_light_flash", False),
        ("DOOR_LOCK", "trigger_remote_door_lock", True),
        ("DOOR_UNLOCK", "trigger_remote_door_unlock", True),
        ("CLIMATE_NOW", "trigger_remote_air_conditioning", True),
        ("CLIMATE_STOP", "trigger_remote_air_conditioning_stop", True),
        ("VEHICLE_FINDER", "trigger_remote_vehicle_finder", False),
        ("HORN_BLOW", "trigger_remote_horn", False),
        ("SEND_POI", "trigger_send_poi", False),
        ("CHARGE_NOW", "trigger_charge_now", True),
    ]

    account = await get_mocked_account()
    vehicle = account.get_vehicle(VIN_F45)

    for service, call, triggers_update in services:
        with mock.patch(
            "bimmer_connected.account.MyBMWAccount.get_vehicles", new_callable=mock.AsyncMock
        ) as mock_listener:
            mock_listener.reset_mock()

            if service == "SEND_POI":
                response = await getattr(vehicle.remote_services, call)(POI_DATA)
            else:
                response = await getattr(vehicle.remote_services, call)()
                assert ExecutionState.EXECUTED == response.state

                if triggers_update:
                    mock_listener.assert_called_once_with()
                else:
                    mock_listener.assert_not_called()


@pytest.mark.asyncio
async def test_get_remote_service_status():
    """Test get_remove_service_status method."""
    # pylint: disable=protected-access
    remote_services._POLLING_CYCLE = 0

    account = await get_mocked_account()
    vehicle = account.get_vehicle(VIN_F45)

    with remote_services_mock() as mock_api:
        mock_api.post("/eadrax-vrccs/v2/presentation/remote-commands/eventStatus", params={"eventId": mock.ANY}).mock(
            side_effect=[
                httpx.Response(500),
                httpx.Response(200, text="You can't parse this..."),
                httpx.Response(200, json=load_response(_RESPONSE_ERROR)),
            ],
        )

        with pytest.raises(httpx.HTTPStatusError):
            await vehicle.remote_services._block_until_done(uuid4())
        with pytest.raises(ValueError):
            await vehicle.remote_services._block_until_done(uuid4())
        with pytest.raises(Exception):
            await vehicle.remote_services._block_until_done(uuid4())


@remote_services_mock()
@pytest.mark.asyncio
async def test_get_remote_position():
    """Test getting position from remote service."""
    remote_services._POLLING_CYCLE = 0  # pylint: disable=protected-access

    account = await get_mocked_account()
    account.set_observer_position(1.0, 0.0)
    vehicle = account.get_vehicle(VIN_F45)
    status = vehicle.status

    # Check original position
    assert (12.3456, 34.5678) == status.gps_position
    assert 123 == status.gps_heading

    # Check updated position
    await vehicle.remote_services.trigger_remote_vehicle_finder()
    assert (123.456, 34.5678) == status.gps_position
    assert 121 == status.gps_heading

    # Position should still be from vehicle finder after status update
    await account.get_vehicles()
    assert (123.456, 34.5678) == status.gps_position
    assert 121 == status.gps_heading


@remote_services_mock()
@pytest.mark.asyncio
async def test_get_remote_position_fail_without_observer(caplog):
    """Test getting position from remote service."""
    remote_services._POLLING_CYCLE = 0  # pylint: disable=protected-access

    account = await get_mocked_account()
    vehicle = account.get_vehicle(VIN_F45)

    await vehicle.remote_services.trigger_remote_vehicle_finder()
    errors = [
        r
        for r in caplog.records
        if r.levelname == "ERROR"
        and "Unknown position: Set observer position to retrieve vehicle coordinates" in r.message
    ]
    assert len(errors) == 1


@remote_services_mock()
@pytest.mark.asyncio
async def test_fail_with_timeout():
    """Test failing after timeout was reached."""
    remote_services._POLLING_CYCLE = 1  # pylint: disable=protected-access
    remote_services._POLLING_TIMEOUT = 2  # pylint: disable=protected-access

    account = await get_mocked_account()
    vehicle = account.get_vehicle(VIN_F45)

    with pytest.raises(TimeoutError):
        await vehicle.remote_services.trigger_remote_light_flash()


@time_machine.travel("2020-01-01", tick=False)
@remote_services_mock()
@pytest.mark.asyncio
async def test_get_remote_position_too_old():
    """Test remote service position being ignored as vehicle status is newer."""
    remote_services._POLLING_CYCLE = 0  # pylint: disable=protected-access

    account = await get_mocked_account()
    vehicle = account.get_vehicle(VIN_F45)
    status = vehicle.status

    await vehicle.remote_services.trigger_remote_vehicle_finder()

    assert (12.3456, 34.5678) == status.gps_position
    assert 123 == status.gps_heading


@remote_services_mock()
@pytest.mark.asyncio
async def test_poi():
    """Test get_remove_service_status method."""
    remote_services._POLLING_CYCLE = 0  # pylint: disable=protected-access

    account = await get_mocked_account()
    vehicle = account.get_vehicle(VIN_F45)

    with pytest.raises(TypeError):
        await vehicle.remote_services.trigger_send_poi({"lat": 12.34})



"""Tests for API that are not covered by other tests."""
import json

import pytest

from bimmer_connected.api.regions import get_region_from_name, valid_regions
from bimmer_connected.api.utils import anonymize_data, log_to_to_file


def test_valid_regions():
    """Test valid regions."""
    assert ["north_america", "china", "rest_of_world"] == valid_regions()


def test_unknown_region():
    """Test unknown region."""
    with pytest.raises(ValueError):
        get_region_from_name("unknown")


def test_anonymize_data():
    """Test anonymization function."""
    test_dict = {
        "vin": "secret",
        "a sub-dict": {
            "lat": 666,
            "lon": 666,
            "heading": 666,
        },
        "licensePlate": "secret",
        "public": "public_data",
        "a_list": [
            {"vin": "secret"},
            {
                "lon": 666,
                "public": "more_public_data",
            },
        ],
        "b_list": ["a", "b"],
        "empty_list": [],
    }
    anon_text = json.dumps(anonymize_data(test_dict))
    assert "secret" not in anon_text
    assert "666" not in anon_text
    assert "public_data" in anon_text
    assert "more_public_data" in anon_text


def test_log_to_file_without_file_name(tmp_path):
    """Test not logging to file if no file name is given."""
    assert log_to_to_file(content=[], logfile_path=tmp_path, logfile_name=None) is None
    assert len(list(tmp_path.iterdir())) == 0



"""Authentication management for BMW APIs."""

import asyncio
import base64
import datetime
import logging
from dataclasses import dataclass
from typing import Optional

import httpx
import jwt
from Crypto.Cipher import PKCS1_v1_5
from Crypto.PublicKey import RSA

from bimmer_connected.api.regions import Regions, get_ocp_apim_key, get_server_url
from bimmer_connected.api.utils import (
    create_s256_code_challenge,
    generate_token,
    handle_http_status_error,
    raise_for_status_event_handler,
)
from bimmer_connected.const import AUTH_CHINA_LOGIN_URL, AUTH_CHINA_PUBLIC_KEY_URL, OAUTH_CONFIG_URL, X_USER_AGENT

EXPIRES_AT_OFFSET = datetime.timedelta(seconds=10)

_LOGGER = logging.getLogger(__name__)


@dataclass
class Authentication:
    """Base class for Authentication."""

    username: str
    password: str
    region: Regions
    token: Optional[str] = None
    expires_at: Optional[datetime.datetime] = None

    async def login(self) -> None:
        """Get a valid OAuth token."""
        raise NotImplementedError("Not implemented in Authentication base class.")

    async def get_authentication(self) -> str:
        """Returns a valid Bearer token."""
        if not self.is_token_valid:
            await self.login()
        return f"Bearer {self.token}"

    @property
    def is_token_valid(self) -> bool:
        """Check if current token is still valid."""
        if self.token and self.expires_at and datetime.datetime.utcnow() < self.expires_at:
            _LOGGER.debug("Old token is still valid. Not getting a new one.")
            return True
        return False


@dataclass
class MyBMWAuthentication(Authentication):
    """Authentication for MyBMW API."""

    _lock: Optional[asyncio.Lock] = None

    def _create_or_update_lock(self):
        """Makes sure that there is a lock in the current event loop."""
        if not self._lock:
            self._lock = asyncio.Lock()

    async def login(self) -> None:
        """Get a valid OAuth token."""
        self._create_or_update_lock()
        async with self._lock:
            if self.is_token_valid:
                return

            token_data = {}
            if self.region in [Regions.NORTH_AMERICA, Regions.REST_OF_WORLD]:
                token_data = await self._login_row_na()
            elif self.region in [Regions.CHINA]:
                token_data = await self._login_china()

            self.token = token_data["access_token"]
            self.expires_at = token_data["expires_at"] - EXPIRES_AT_OFFSET

    async def _login_row_na(self):  # pylint: disable=too-many-locals
        """Login to Rest of World and North America."""
        try:
            async with httpx.AsyncClient(
                base_url=get_server_url(self.region), headers={"user-agent": "Dart/2.13 (dart:io)"}
            ) as client:
                _LOGGER.debug("Authenticating with MyBMW flow for North America & Rest of World.")

                # Attach raise_for_status event hook
                client.event_hooks["response"].append(raise_for_status_event_handler)

                # Get OAuth2 settings from BMW API
                r_oauth_settings = await client.get(
                    OAUTH_CONFIG_URL,
                    headers={
                        "ocp-apim-subscription-key": get_ocp_apim_key(self.region),
                        "x-user-agent": X_USER_AGENT.format("bmw"),
                    },
                )
                oauth_settings = r_oauth_settings.json()

                # Generate OAuth2 Code Challenge + State
                code_verifier = generate_token(86)
                code_challenge = create_s256_code_challenge(code_verifier)

                state = generate_token(22)

                # Set up authenticate endpoint
                authenticate_url = oauth_settings["tokenEndpoint"].replace("/token", "/authenticate")
                oauth_base_values = {
                    "client_id": oauth_settings["clientId"],
                    "response_type": "code",
                    "redirect_uri": oauth_settings["returnUrl"],
                    "state": state,
                    "nonce": "login_nonce",
                    "scope": " ".join(oauth_settings["scopes"]),
                    "code_challenge": code_challenge,
                    "code_challenge_method": "S256",
                }

                # Call authenticate endpoint first time (with user/pw) and get authentication
                response = await client.post(
                    authenticate_url,
                    data=dict(
                        oauth_base_values,
                        **{
                            "grant_type": "authorization_code",
                            "username": self.username,
                            "password": self.password,
                        },
                    ),
                )
                authorization = httpx.URL(response.json()["redirect_to"]).params["authorization"]

                # With authorization, call authenticate endpoint second time to get code
                response = await client.post(
                    authenticate_url,
                    data=dict(oauth_base_values, **{"authorization": authorization}),
                )
                code = response.next_request.url.params["code"]

                # With code, get token
                current_utc_time = datetime.datetime.utcnow()
                response = await client.post(
                    oauth_settings["tokenEndpoint"],
                    data={
                        "code": code,
                        "code_verifier": code_verifier,
                        "redirect_uri": oauth_settings["returnUrl"],
                        "grant_type": "authorization_code",
                    },
                    auth=(oauth_settings["clientId"], oauth_settings["clientSecret"]),
                )
                response_json = response.json()

                expiration_time = int(response_json["expires_in"])
                expires_at = current_utc_time + datetime.timedelta(seconds=expiration_time)

        except httpx.HTTPStatusError as ex:
            handle_http_status_error(ex, "Authentication", _LOGGER)

        return {"access_token": response_json["access_token"], "expires_at": expires_at}

    async def _login_china(self):
        try:
            async with httpx.AsyncClient(
                base_url=get_server_url(self.region), headers={"user-agent": "Dart/2.13 (dart:io)"}
            ) as client:
                _LOGGER.debug("Authenticating with MyBMW flow for China.")

                # Attach raise_for_status event hook
                client.event_hooks["response"].append(raise_for_status_event_handler)

                login_header = {"x-user-agent": X_USER_AGENT.format("bmw")}

                # Get current RSA public certificate & use it to encrypt password
                response = await client.get(
                    AUTH_CHINA_PUBLIC_KEY_URL,
                    headers=login_header,
                )
                pem_public_key = response.json()["data"]["value"]

                public_key = RSA.import_key(pem_public_key)
                cipher_rsa = PKCS1_v1_5.new(public_key)
                encrypted = cipher_rsa.encrypt(self.password.encode())
                pw_encrypted = base64.b64encode(encrypted).decode("UTF-8")

                # Get token
                response = await client.post(
                    AUTH_CHINA_LOGIN_URL, headers=login_header, json={"mobile": self.username, "password": pw_encrypted}
                )
                response_json = response.json()["data"]

                decoded_token = jwt.decode(
                    response_json["access_token"], algorithms=["HS256"], options={"verify_signature": False}
                )

        except httpx.HTTPStatusError as ex:
            handle_http_status_error(ex, "Authentication", _LOGGER)

        return {
            "access_token": response_json["access_token"],
            "expires_at": datetime.datetime.utcfromtimestamp(decoded_token["exp"]),
        }



"""Library to read data from the BMW Connected Drive portal.

The library bimmer_connected provides a Python interface to interact
with the BMW Connected Drive web service. It allows you to read
the current state of the vehicle and also trigger remote services.

Disclaimer:
This library is not affiliated with or endorsed by BMW Group.
"""

import pkg_resources  # part of setuptools

__version__ = pkg_resources.get_distribution("bimmer_connected").version



"""General utils and base classes used in the library."""

import datetime
import inspect
import json
import logging
import sys
import traceback
from enum import Enum
from typing import TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:
    from typing import Callable, TypeVar

    from typing_extensions import Concatenate, ParamSpec

    _T = TypeVar("_T")
    _R = TypeVar("_R")
    _P = ParamSpec("_P")
_LOGGER = logging.getLogger(__name__)


JSON_IGNORED_KEYS = ["account", "_account", "vehicle", "_vehicle", "status", "remote_services"]
JSON_DEPRECATED_KEYS = [
    "has_hv_battery",
    "has_range_extender",
    "has_internal_combustion_engine",
    "has_weekly_planner_service",
]


def get_class_property_names(obj: object):
    """Returns the names of all properties of a class."""
    return [p[0] for p in inspect.getmembers(type(obj), inspect.isdatadescriptor) if not p[0].startswith("_")]


def parse_datetime(date_str: str) -> Optional[datetime.datetime]:
    """Convert a time string into datetime."""
    if not date_str:
        return None
    date_formats = ["%Y-%m-%dT%H:%M:%S.%f%z", "%Y-%m-%dT%H:%M:%S%z", "%Y-%m-%dT%H:%M:%S.%fZ", "%Y-%m-%dT%H:%M:%SZ"]
    for date_format in date_formats:
        try:
            parsed = datetime.datetime.strptime(date_str, date_format)
            # Assume implicit UTC for Python 3.6
            if sys.version_info < (3, 7):
                parsed = parsed.replace(tzinfo=datetime.timezone.utc)
            return parsed
        except ValueError:
            pass
    _LOGGER.error("unable to parse '%s' using %s", date_str, date_formats)
    return None


class MyBMWJSONEncoder(json.JSONEncoder):
    """JSON Encoder that handles data classes, properties and additional data types."""

    def default(self, o):
        if isinstance(o, (datetime.datetime, datetime.date, datetime.time)):
            return o.isoformat()
        if not isinstance(o, Enum) and hasattr(o, "__dict__") and isinstance(o.__dict__, Dict):
            retval: Dict = o.__dict__
            retval.update({p: getattr(o, p) for p in get_class_property_names(o) if p not in JSON_DEPRECATED_KEYS})
            return {k: v for k, v in retval.items() if k not in JSON_IGNORED_KEYS}
        return str(o)


def deprecated(replacement: str = None):
    """Mark a function or property as deprecated."""

    def decorator(func: "Callable[Concatenate[_T, _P], _R]") -> "Callable[Concatenate[_T, _P], _R | None]":
        def _func_wrapper(self: "_T", *args: "_P.args", **kwargs: "_P.kwargs") -> "_R | None":
            # warnings.simplefilter('always', DeprecationWarning)  # turn off filter
            replacement_text = f" Please change to '{replacement}'." if replacement else ""
            # warnings.warn(f"{func.__qualname__} is deprecated.{replacement_text}",
            # category=DeprecationWarning,
            # stacklevel=2)
            # warnings.simplefilter('default', DeprecationWarning)  # reset filter
            stack = traceback.extract_stack()[-2]
            _LOGGER.warning(
                "DeprecationWarning:%s:%s: '%s' is deprecated.%s",
                stack.filename,
                stack.lineno,
                func.__qualname__,
                replacement_text,
            )
            return func(self, *args, **kwargs)

        return _func_wrapper

    return decorator



"""Tests for deprecated MyBMWVehicle."""
import pytest

from bimmer_connected.const import CarBrands
from bimmer_connected.vehicle.vehicle import ConnectedDriveVehicle

from . import (
    VIN_F11,
    VIN_F31,
    VIN_F35,
    VIN_F44,
    VIN_F45,
    VIN_F48,
    VIN_G01,
    VIN_G05,
    VIN_G08,
    VIN_G21,
    VIN_G30,
    VIN_I01_NOREX,
    VIN_I01_REX,
    get_deprecation_warning_count,
)
from .test_account import get_mocked_account

ATTRIBUTE_MAPPING = {
    "remainingFuel": "remaining_fuel",
    "position": "gps_position",
    "cbsData": "condition_based_services",
    "checkControlMessages": "check_control_messages",
    "doorLockState": "door_lock_state",
    "updateReason": "last_update_reason",
    "chargingLevelHv": "charging_level_hv",
    "chargingStatus": "charging_status",
    "maxRangeElectric": "max_range_electric",
    "remainingRangeElectric": "remaining_range_electric",
    "parkingLight": "parking_lights",
    "remainingRangeFuel": "remaining_range_fuel",
    "updateTime": "timestamp",
    "chargingTimeRemaining": "charging_time_remaining",
}


@pytest.mark.asyncio
async def test_parsing_attributes(caplog):
    """Test parsing different attributes of the vehicle."""
    account = await get_mocked_account()

    for vehicle in account.vehicles:
        print(vehicle.name)
        assert vehicle.drive_train is not None
        assert vehicle.name is not None
        assert isinstance(vehicle.brand, CarBrands)
        assert vehicle.has_internal_combustion_engine is not None  #
        assert vehicle.has_hv_battery is not None  #
        assert vehicle.drive_train_attributes is not None
        assert vehicle.has_weekly_planner_service is not None  #

    assert len(get_deprecation_warning_count(caplog)) == len(account.vehicles) * 3


@pytest.mark.asyncio
async def test_drive_train_attributes(caplog):
    """Test parsing different attributes of the vehicle."""
    account = await get_mocked_account()

    vehicle_drivetrains = {
        VIN_F11: (True, False, False),
        VIN_F31: (True, False, False),
        VIN_F35: (True, False, False),
        VIN_F44: (True, False, False),
        VIN_F45: (True, True, False),
        VIN_F48: (True, False, False),
        VIN_G01: (True, True, False),
        VIN_G05: (True, True, False),
        VIN_G08: (False, True, False),
        VIN_G21: (True, True, False),
        VIN_G30: (True, True, False),
        VIN_I01_NOREX: (False, True, False),
        VIN_I01_REX: (False, True, True),
    }

    for vehicle in account.vehicles:
        assert vehicle_drivetrains[vehicle.vin][0] == vehicle.has_internal_combustion_engine
        assert vehicle_drivetrains[vehicle.vin][1] == vehicle.has_hv_battery
        assert vehicle_drivetrains[vehicle.vin][2] == vehicle.has_range_extender

    assert len(get_deprecation_warning_count(caplog)) == len(account.vehicles) * 3


@pytest.mark.asyncio
async def test_deprecated_vehicle(caplog):
    """Test deprecation warning for ConnectedDriveVehicle."""
    account = await get_mocked_account()

    deprecated_vehicle = ConnectedDriveVehicle(account, account.vehicles[0].data)

    assert deprecated_vehicle is not None
    assert len(get_deprecation_warning_count(caplog)) == 1



"""Generals models used for bimmer_connected."""

import logging
from dataclasses import InitVar, dataclass, field
from enum import Enum
from typing import Dict, NamedTuple, Optional, Tuple, Union

_LOGGER = logging.getLogger(__name__)


class StrEnum(str, Enum):
    """A string enumeration of type `(str, Enum)`. All members are compared via `upper()`."""

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value.upper() == value.upper():
                return member
        raise ValueError(f"'{value}' is not a valid {cls.__name__}")


@dataclass
class VehicleDataBase:
    """A base class for parsing and storing complex vehicle data."""

    @classmethod
    def from_vehicle_data(cls, vehicle_data: Dict):
        """Creates the class based on vehicle data from API."""
        parsed = cls._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return cls(**parsed)
        return None

    def update_from_vehicle_data(self, vehicle_data: Dict):
        """Updates the attributes based on vehicle data from API."""
        parsed = self._parse_vehicle_data(vehicle_data) or {}
        parsed.update(self._update_after_parse(parsed))
        if len(parsed) > 0:
            self.__dict__.update(parsed)

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Optional[Dict]:
        """Parses desired attributes out of vehicle data from API."""
        raise NotImplementedError()

    def _update_after_parse(self, parsed: Dict) -> Dict:
        """Updates parsed vehicle data with attributes stored in class if needed."""
        # pylint:disable=no-self-use
        return parsed


@dataclass
class GPSPosition:
    """GPS coordinates."""

    latitude: Optional[float]
    longitude: Optional[float]

    def __post_init__(self):
        # pylint: disable=no-member
        if len([v for v in self.__dict__.values() if v is None]) not in [0, len(self.__dataclass_fields__)]:
            raise TypeError("Either none or all arguments must be 'None'.")

        # pylint: disable=no-member
        for field_name in self.__dataclass_fields__:
            value = getattr(self, field_name)
            if value is not None and not isinstance(value, (float, int)):
                raise TypeError(f"'{field_name}' not of type '{Optional[Union[float, int]]}'")

    def __iter__(self):
        yield from self.__dict__.values()

    def __getitem__(self, key):
        return tuple(self.__dict__.values())[key]

    def __eq__(self, other):
        if isinstance(other, Tuple):
            return tuple(self.__iter__()) == other
        if hasattr(self, "__dict__") and hasattr(other, "__dict__"):
            return self.__dict__ == other.__dict__
        return False


@dataclass
class PointOfInterestAddress:
    """Address data of a PointOfInterest."""

    street: Optional[str] = None
    postalCode: Optional[str] = None  # pylint: disable=invalid-name
    city: Optional[str] = None
    country: Optional[str] = None


@dataclass
class PointOfInterest:
    """A Point of Interest to be sent to the car."""

    lat: InitVar[float]
    lon: InitVar[float]
    name: Optional[str] = None
    street: InitVar[str] = None
    postal_code: InitVar[str] = None
    city: InitVar[str] = None
    country: InitVar[str] = None

    coordinates: GPSPosition = field(init=False)
    locationAddress: Optional[PointOfInterestAddress] = field(init=False)  # pylint: disable=invalid-name
    type: str = field(default="SHARED_DESTINATION_FROM_EXTERNAL_APP", init=False)

    def __post_init__(self, lat, lon, street, postal_code, city, country):  # pylint: disable=too-many-arguments
        self.coordinates = GPSPosition(lat, lon)
        # pylint: disable=invalid-name
        self.locationAddress = PointOfInterestAddress(street, postal_code, city, country)


def check_strict_types(cls):
    """Checks a dataclass for strict typing. Use in __post_init__."""
    for field_name, field_def in cls.__dataclass_fields__.items():  # pylint: disable=no-member
        try:
            original_type = field_def.type.__args__
            field_type = original_type or field_def.type
        except AttributeError:
            field_type = field_def.type

        if not isinstance(getattr(cls, field_name), field_type):
            raise TypeError(f"'{field_name}' not of type '{field_def.type}'")


class ValueWithUnit(NamedTuple):
    """A value with a corresponding unit."""

    value: Optional[Union[int, float]]
    unit: Optional[str]



"""URLs for different services and error code mapping."""
from enum import Enum


class CarBrands(str, Enum):
    """Car brands supported by the MyBMW API."""

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value == value.lower():
                return member
        raise ValueError(f"'{value}' is not a valid {cls.__name__}")

    BMW = "bmw"
    MINI = "mini"


class Regions(Enum):
    """Regions of the world with separate servers."""

    NORTH_AMERICA = 0
    CHINA = 1
    REST_OF_WORLD = 2


SERVER_URLS_MYBMW = {
    Regions.NORTH_AMERICA: "cocoapi.bmwgroup.us",
    Regions.REST_OF_WORLD: "cocoapi.bmwgroup.com",
    Regions.CHINA: "myprofile.bmw.com.cn",
}

OCP_APIM_KEYS = {
    Regions.NORTH_AMERICA: "31e102f5-6f7e-7ef3-9044-ddce63891362",
    Regions.REST_OF_WORLD: "4f1c85a3-758f-a37d-bbb6-f8704494acfa",
}

X_USER_AGENT = "android(v1.07_20200330);{};1.7.0(11152)"

AUTH_CHINA_PUBLIC_KEY_URL = "/eadrax-coas/v1/cop/publickey"
AUTH_CHINA_LOGIN_URL = "/eadrax-coas/v1/login/pwd"

OAUTH_CONFIG_URL = "/eadrax-ucs/v1/presentation/oauth/config"

VEHICLES_URL = "/eadrax-vcs/v1/vehicles"

REMOTE_SERVICE_BASE_URL = "/eadrax-vrccs/v2/presentation/remote-commands"
REMOTE_SERVICE_URL = REMOTE_SERVICE_BASE_URL + "/{vin}/{service_type}"
REMOTE_SERVICE_STATUS_URL = REMOTE_SERVICE_BASE_URL + "/eventStatus?eventId={event_id}"
REMOTE_SERVICE_POSITION_URL = REMOTE_SERVICE_BASE_URL + "/eventPosition?eventId={event_id}"

VEHICLE_IMAGE_URL = "/eadrax-ics/v3/presentation/vehicles/{vin}/images?carView={view}"
VEHICLE_POI_URL = "/eadrax-dcs/v1/send-to-car/send-to-car"

VEHICLE_CHARGING_STATISTICS_URL = "/eadrax-chs/v1/charging-statistics"
VEHICLE_CHARGING_SESSIONS_URL = "/eadrax-chs/v1/charging-sessions"

SERVICE_PROPERTIES = "properties"
SERVICE_STATUS = "status"
SERVICE_CHARGING_STATISTICS_URL = "CHARGING_STATISTICS"
SERVICE_CHARGING_SESSIONS_URL = "CHARGING_SESSIONS"
SERVICE_CHARGING_PROFILE = "CHARGING_PROFILE"



"""Test for VehicleState."""

import datetime

import pytest
import time_machine

from bimmer_connected.api.regions import get_region_from_name
from bimmer_connected.vehicle.doors_windows import LidState, LockState
from bimmer_connected.vehicle.fuel_and_battery import ChargingState
from bimmer_connected.vehicle.reports import CheckControlStatus, ConditionBasedServiceStatus

from . import VIN_F11, VIN_F31, VIN_F48, VIN_G01, VIN_G08, VIN_G30, VIN_I01_REX, get_deprecation_warning_count
from .test_account import get_mocked_account


@pytest.mark.asyncio
async def test_generic(caplog):
    """Test generic attributes."""
    status = (await get_mocked_account()).get_vehicle(VIN_G30)

    expected = datetime.datetime(
        year=2021, month=11, day=11, hour=8, minute=58, second=53, tzinfo=datetime.timezone.utc
    )
    assert expected == status.timestamp

    assert 7991 == status.mileage[0]
    assert "km" == status.mileage[1]

    assert (12.3456, 34.5678) == status.vehicle_location.location
    assert 123 == status.vehicle_location.heading

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_range_combustion_no_info(caplog):
    """Test if the parsing of mileage and range is working"""
    status = (await get_mocked_account()).get_vehicle(VIN_F31).fuel_and_battery

    assert (32, "LITERS") == status.remaining_fuel
    assert status.remaining_range_fuel == (None, None)
    assert status.remaining_fuel_percent is None

    assert status.remaining_battery_percent is None
    assert status.remaining_range_electric == (None, None)

    assert status.remaining_range_total == (None, None)

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_range_combustion(caplog):
    """Test if the parsing of mileage and range is working"""
    status = (await get_mocked_account()).get_vehicle(VIN_F48).fuel_and_battery

    assert (19, "LITERS") == status.remaining_fuel
    assert (308, "km") == status.remaining_range_fuel
    assert status.remaining_fuel_percent is None

    assert status.remaining_battery_percent is None
    assert status.remaining_range_electric == (None, None)

    assert (308, "km") == status.remaining_range_total

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_range_phev(caplog):
    """Test if the parsing of mileage and range is working"""
    status = (await get_mocked_account()).get_vehicle(VIN_G30).fuel_and_battery

    assert (11, "LITERS") == status.remaining_fuel
    assert (107, "km") == status.remaining_range_fuel
    assert 28 == status.remaining_fuel_percent

    assert 41 == status.remaining_battery_percent
    assert (9, "km") == status.remaining_range_electric

    assert (116, "km") == status.remaining_range_total

    assert status.remaining_range_fuel[0] + status.remaining_range_electric[0] == status.remaining_range_total[0]

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_range_rex(caplog):
    """Test if the parsing of mileage and range is working"""
    status = (await get_mocked_account()).get_vehicle(VIN_I01_REX).fuel_and_battery

    assert (5, "LITERS") == status.remaining_fuel
    assert (64, "km") == status.remaining_range_fuel
    assert status.remaining_fuel_percent is None

    assert 100 == status.remaining_battery_percent
    assert (164, "km") == status.remaining_range_electric

    assert (228, "km") == status.remaining_range_total

    assert status.remaining_range_fuel[0] + status.remaining_range_electric[0] == status.remaining_range_total[0]

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_range_electric(caplog):
    """Test if the parsing of mileage and range is working"""
    status = (await get_mocked_account()).get_vehicle(VIN_G08).fuel_and_battery

    assert (0, "LITERS") == status.remaining_fuel
    assert status.remaining_range_fuel == (None, None)
    assert status.remaining_fuel_percent is None

    assert 50 == status.remaining_battery_percent
    assert (179, "km") == status.remaining_range_electric

    assert (179, "km") == status.remaining_range_total

    assert len(get_deprecation_warning_count(caplog)) == 0


@time_machine.travel("2011-11-28 21:28:59 +0000", tick=False)
@pytest.mark.asyncio
async def test_charging_end_time(caplog):
    """Test if the parsing of mileage and range is working"""
    account = await get_mocked_account()
    status = account.get_vehicle(VIN_G08).fuel_and_battery
    assert datetime.datetime(2011, 11, 29, 4, 1, tzinfo=account.timezone) == status.charging_end_time

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_charging_time_label(caplog):
    """Test if the parsing of mileage and range is working"""
    account = await get_mocked_account()
    status = account.get_vehicle(VIN_G08).fuel_and_battery
    assert "100% at ~04:01 AM" == status.charging_time_label

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_charging_end_time_parsing_failure(caplog):
    """Test if the parsing of mileage and range is working"""
    account = await get_mocked_account()
    vehicle = account.get_vehicle(VIN_G08)

    vehicle.update_state(
        dict(
            vehicle.data,
            **{
                "status": {
                    "fuelIndicators": [
                        {
                            "chargingStatusIndicatorType": "CHARGING",
                            "chargingStatusType": "CHARGING",
                            "infoLabel": "100% at later today...",
                            "rangeIconId": 59683,
                            "rangeUnits": "km",
                            "rangeValue": "179",
                        }
                    ]
                },
            },
        )
    )
    assert vehicle.fuel_and_battery.charging_end_time is None
    assert "100% at later today..." == vehicle.fuel_and_battery.charging_time_label

    errors = [r for r in caplog.records if r.levelname == "ERROR" and "Error parsing charging end time" in r.message]
    assert len(errors) == 1

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_plugged_in_waiting_for_charge_window(caplog):
    """G01 is plugged in but not charging, as its waiting for charging window."""
    # Should be None on G01 as it is only "charging"
    account = await get_mocked_account()
    vehicle = account.get_vehicle(VIN_G01)

    assert vehicle.fuel_and_battery.charging_end_time is None
    assert "Starts at ~ 09:00 AM" == vehicle.fuel_and_battery.charging_time_label
    assert ChargingState.PLUGGED_IN == vehicle.fuel_and_battery.charging_status
    assert vehicle.fuel_and_battery.is_charger_connected is True

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_condition_based_services(caplog):
    """Test condition based service messages."""
    vehicle = (await get_mocked_account()).get_vehicle(VIN_G30)

    cbs = vehicle.condition_based_services.messages
    assert 3 == len(cbs)
    assert ConditionBasedServiceStatus.OK == cbs[0].state
    expected_cbs0 = datetime.datetime(year=2022, month=8, day=1, tzinfo=datetime.timezone.utc)
    assert expected_cbs0 == cbs[0].due_date
    assert (25000, "KILOMETERS") == cbs[0].due_distance

    assert ConditionBasedServiceStatus.OK == cbs[1].state
    expected_cbs1 = datetime.datetime(year=2023, month=8, day=1, tzinfo=datetime.timezone.utc)
    assert expected_cbs1 == cbs[1].due_date
    assert (None, None) == cbs[1].due_distance

    assert ConditionBasedServiceStatus.OK == cbs[2].state
    expected_cbs2 = datetime.datetime(year=2024, month=8, day=1, tzinfo=datetime.timezone.utc)
    assert expected_cbs2 == cbs[2].due_date
    assert (60000, "KILOMETERS") == cbs[2].due_distance

    assert vehicle.condition_based_services.is_service_required is False

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_parse_f31_no_position(caplog):
    """Test parsing of F31 data with position tracking disabled in the vehicle."""
    vehicle = (await get_mocked_account()).get_vehicle(VIN_F31)

    assert vehicle.vehicle_location.location == (None, None)
    assert vehicle.vehicle_location.heading is None

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_parse_gcj02_position(caplog):
    """Test conversion of GCJ02 to WGS84 for china."""
    account = await get_mocked_account(get_region_from_name("china"))
    vehicle = account.get_vehicle(VIN_F48)

    vehicle_test_data = {
        "properties": {
            "vehicleLocation": {
                "address": {"formatted": "some_formatted_address"},
                "coordinates": {"latitude": 39.83492, "longitude": 116.23221},
                "heading": 123,
            },
            "lastUpdatedAt": "2021-11-14T20:20:21Z",
        },
        "status": {
            "FuelAndBattery": [],
            "lastUpdatedAt": "2021-11-14T20:20:21Z",
        },
    }

    vehicle.update_state(dict(vehicle.data, **vehicle_test_data))

    # Update twice to test against slowly crawling position due to GCJ02 to WGS84 conversion
    vehicle.update_state(dict(vehicle.data, **vehicle_test_data))

    assert (39.8337, 116.22617) == (
        round(vehicle.vehicle_location.location[0], 5),
        round(vehicle.vehicle_location.location[1], 5),
    )

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_parse_g08(caplog):
    """Test if the parsing of the attributes is working."""
    status = (await get_mocked_account()).get_vehicle(VIN_G08).fuel_and_battery

    assert (179, "km") == status.remaining_range_electric
    assert (179, "km") == status.remaining_range_total
    assert ChargingState.CHARGING == status.charging_status
    assert 50 == status.remaining_battery_percent

    assert len(get_deprecation_warning_count(caplog)) == 0


# def test_missing_attribute(caplog):
#     """Test if error handling is working correctly."""
#     account = unittest.mock.MagicMock(MyBMWAccount)
#     state = VehicleState(account, None)
#     state._attributes[SERVICE_STATUS] = {}
#     assert status.mileage is None


@pytest.mark.asyncio
async def test_lids(caplog):
    """Test features around lids."""
    status = (await get_mocked_account()).get_vehicle(VIN_G30).doors_and_windows

    assert 6 == len(list(status.lids))
    assert 3 == len(list(status.open_lids))
    assert status.all_lids_closed is False

    status = (await get_mocked_account()).get_vehicle(VIN_G08).doors_and_windows

    for lid in status.lids:
        assert LidState.CLOSED == lid.state
    assert status.all_lids_closed is True
    assert 6 == len(list(status.lids))

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_windows_g31(caplog):
    """Test features around windows."""
    status = (await get_mocked_account()).get_vehicle(VIN_G08).doors_and_windows

    for window in status.windows:
        assert LidState.CLOSED == window.state

    assert 5 == len(list(status.windows))
    assert 0 == len(list(status.open_windows))
    assert status.all_windows_closed is True

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_door_locks(caplog):
    """Test the door locks."""
    status = (await get_mocked_account()).get_vehicle(VIN_G08).doors_and_windows

    assert LockState.LOCKED == status.door_lock_state

    status = (await get_mocked_account()).get_vehicle(VIN_I01_REX).doors_and_windows

    assert LockState.UNLOCKED == status.door_lock_state

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_check_control_messages(caplog):
    """Test handling of check control messages.

    F11 is the only vehicle with active Check Control Messages, so we only expect to get something there.
    However we have no vehicle with issues in check control.
    """
    vehicle = (await get_mocked_account()).get_vehicle(VIN_F11)
    assert vehicle.check_control_messages.has_check_control_messages is True

    ccms = vehicle.check_control_messages.messages
    assert 2 == len(ccms)

    assert CheckControlStatus.MEDIUM == ccms[0].state
    assert (
        "Charge by driving for longer periods or use external charger. "
        "Functions requiring battery will be switched off."
    ) == ccms[0].description_long

    assert "Battery discharged: Start engine" == ccms[0].description_short

    assert CheckControlStatus.LOW == ccms[1].state
    assert (
        "System unable to monitor tire pressure. Check tire pressures manually. "
        "Continued driving possible. Consult service center."
    ) == ccms[1].description_long

    assert "Flat Tire Monitor (FTM) inactive" == ccms[1].description_short

    assert len(get_deprecation_warning_count(caplog)) == 0



"""Models the state of a vehicle."""

import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List

from bimmer_connected.vehicle.models import StrEnum, VehicleDataBase

_LOGGER = logging.getLogger(__name__)


class LidState(StrEnum):
    """Possible states of the hatch, trunk, doors, windows, sun roof."""

    CLOSED = "CLOSED"
    OPEN = "OPEN"
    OPEN_TILT = "OPEN_TILT"
    INTERMEDIATE = "INTERMEDIATE"
    INVALID = "INVALID"


class LockState(StrEnum):
    """Possible states of the door locks."""

    LOCKED = "LOCKED"
    SECURED = "SECURED"
    SELECTIVE_LOCKED = "SELECTIVE_LOCKED"
    UNLOCKED = "UNLOCKED"
    UNKNOWN = "UNKNOWN"


class Lid:  # pylint: disable=too-few-public-methods
    """A lid of the vehicle.

    Lids are: Doors + Trunk + Hatch
    """

    def __init__(self, name: str, state: str):
        #: name of the lid
        self.name = name
        self.state = LidState(state)

    @property
    def is_closed(self) -> bool:
        """Check if the lid is closed."""
        return self.state == LidState.CLOSED


class Window(Lid):  # pylint: disable=too-few-public-methods,no-member
    """A window of the vehicle.

    A window can be a normal window of the car or the sun roof.
    """


@dataclass
class DoorsAndWindows(VehicleDataBase):  # pylint:disable=too-many-instance-attributes
    """Provides an accessible version of `properties.doorsAndWindows`."""

    door_lock_state: LockState = LockState.UNKNOWN
    """Get state of the door locks."""

    lids: List[Lid] = field(default_factory=list)
    """All lids (doors+hood+trunk) of the car."""

    windows: List[Window] = field(default_factory=list)
    """All windows (doors+sunroof) of the car."""

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Dict:
        """Parse doors and windows."""
        retval: Dict[str, Any] = {}

        if "properties" not in vehicle_data or "doorsAndWindows" not in vehicle_data["status"]:
            _LOGGER.error("Unable to read data from `properties.doorsAndWindows`.")
            return retval

        doors_and_windows = vehicle_data["properties"]["doorsAndWindows"]

        retval["lids"] = [
            Lid(k, v) for k, v in doors_and_windows.items() if k in ["hood", "trunk"] and v != LidState.INVALID
        ] + [Lid(k, v) for k, v in doors_and_windows["doors"].items() if v != LidState.INVALID]

        retval["windows"] = [Window(k, v) for k, v in doors_and_windows["windows"].items() if v != LidState.INVALID]
        if "moonroof" in doors_and_windows:
            retval["windows"].append(Window("moonroof", doors_and_windows["moonroof"]))

        retval["door_lock_state"] = LockState(vehicle_data["status"]["doorsGeneralState"].upper())

        return retval

    @property
    def open_lids(self) -> List[Lid]:
        """Get all open lids of the car."""
        return [lid for lid in self.lids if not lid.is_closed]

    @property
    def all_lids_closed(self) -> bool:
        """Check if all lids are closed."""
        return len(self.open_lids) == 0

    @property
    def open_windows(self) -> List[Window]:
        """Get all open windows of the car."""
        return [lid for lid in self.windows if not lid.is_closed]

    @property
    def all_windows_closed(self) -> bool:
        """Check if all windows are closed."""
        return len(self.open_windows) == 0



"""Tests for MyBMWAccount."""

from typing import Dict, List

import httpx
import pytest
import respx

from bimmer_connected.account import ConnectedDriveAccount, MyBMWAccount
from bimmer_connected.api.authentication import Authentication
from bimmer_connected.api.client import MyBMWClientConfiguration
from bimmer_connected.api.regions import get_region_from_name
from bimmer_connected.vehicle.models import GPSPosition

from . import (
    RESPONSE_DIR,
    TEST_PASSWORD,
    TEST_REGION,
    TEST_REGION_STRING,
    TEST_USERNAME,
    VIN_G21,
    get_deprecation_warning_count,
    get_fingerprint_count,
    load_response,
)


def authenticate_sideeffect(request: httpx.Request) -> httpx.Response:
    """Returns /oauth/authentication response based on request."""
    request_text = request.read().decode("UTF-8")
    if "username" in request_text and "password" in request_text and "grant_type" in request_text:
        return httpx.Response(200, json=load_response(RESPONSE_DIR / "auth" / "authorization_response.json"))
    return httpx.Response(
        302,
        headers={
            "Location": "com.mini.connected://oauth?code=CODE&state=STATE&client_id=CLIENT_ID&nonce=login_nonce",
        },
    )


def vehicles_sideeffect(request: httpx.Request) -> httpx.Response:
    """Returns /vehicles response based on x-user-agent."""
    x_user_agent = request.headers.get("x-user-agent", "").split(";")
    if len(x_user_agent) == 3:
        brand = x_user_agent[1]
    else:
        raise ValueError("x-user-agent not configured correctly!")

    response_vehicles: List[Dict] = []
    files = RESPONSE_DIR.rglob(f"vehicles_v2_{brand}_0.json")
    for file in files:
        response_vehicles.extend(load_response(file))
    return httpx.Response(200, json=response_vehicles)


def account_mock():
    """Returns mocked adapter for auth."""
    router = respx.mock(assert_all_called=False)

    # Login to north_america and rest_of_world
    router.get("/eadrax-ucs/v1/presentation/oauth/config").respond(
        200, json=load_response(RESPONSE_DIR / "auth" / "oauth_config.json")
    )
    router.post("/gcdm/oauth/authenticate").mock(side_effect=authenticate_sideeffect)
    router.post("/gcdm/oauth/token").respond(200, json=load_response(RESPONSE_DIR / "auth" / "auth_token.json"))

    # Login to china
    router.get("/eadrax-coas/v1/cop/publickey").respond(
        200, json=load_response(RESPONSE_DIR / "auth" / "auth_cn_publickey.json")
    )
    router.post("/eadrax-coas/v1/login/pwd").respond(
        200, json=load_response(RESPONSE_DIR / "auth" / "auth_cn_login_pwd.json")
    )

    # Get all vehicle fingerprints
    router.get("/eadrax-vcs/v1/vehicles").mock(side_effect=vehicles_sideeffect)

    return router


def get_account(region=None):
    """Returns account without token and vehicles (sync)."""
    return MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, region or TEST_REGION)


async def get_mocked_account(region=None):
    """Returns pre-mocked account."""
    with account_mock():
        account = get_account(region)
        await account.get_vehicles()
    return account


@account_mock()
@pytest.mark.asyncio
async def test_login_row_na():
    """Test the login flow."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name(TEST_REGION_STRING))
    await account.get_vehicles()
    assert account is not None


@account_mock()
@pytest.mark.asyncio
async def test_login_china():
    """Test raising an error for region `china`."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
    await account.get_vehicles()
    assert account is not None


@account_mock()
@pytest.mark.asyncio
async def test_vehicles():
    """Test the login flow."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
    await account.get_vehicles()

    assert account.mybmw_client_config.authentication.token is not None
    assert get_fingerprint_count() == len(account.vehicles)

    vehicle = account.get_vehicle(VIN_G21)
    assert VIN_G21 == vehicle.vin

    assert account.get_vehicle("invalid_vin") is None


@pytest.mark.asyncio
async def test_invalid_password():
    """Test parsing the results of an invalid password."""
    with account_mock() as mock_api:
        mock_api.post("/gcdm/oauth/authenticate").respond(
            401, json=load_response(RESPONSE_DIR / "auth" / "auth_error_wrong_password.json")
        )
        with pytest.raises(httpx.HTTPStatusError):
            account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
            await account.get_vehicles()


@pytest.mark.asyncio
async def test_invalid_password_china():
    """Test parsing the results of an invalid password."""
    with account_mock() as mock_api:
        mock_api.post("/eadrax-coas/v1/login/pwd").respond(
            422, json=load_response(RESPONSE_DIR / "auth" / "auth_cn_login_error.json")
        )
        with pytest.raises(httpx.HTTPStatusError):
            account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
            await account.get_vehicles()


@pytest.mark.asyncio
async def test_server_error():
    """Test parsing the results of a server error."""
    with account_mock() as mock_api:
        mock_api.post("/gcdm/oauth/authenticate").respond(
            500, text=load_response(RESPONSE_DIR / "auth" / "auth_error_internal_error.txt")
        )
        with pytest.raises(httpx.HTTPStatusError):
            account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
            await account.get_vehicles()


@pytest.mark.asyncio
async def test_vehicle_search_case():
    """Check if the search for the vehicle by VIN is NOT case sensitive."""
    account = await get_mocked_account()

    vin = account.vehicles[1].vin
    assert vin == account.get_vehicle(vin).vin
    assert vin == account.get_vehicle(vin.lower()).vin
    assert vin == account.get_vehicle(vin.upper()).vin


@pytest.mark.asyncio
async def test_storing_fingerprints(tmp_path):
    """Test the login flow."""
    with account_mock() as mock_api:
        account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION, log_responses=tmp_path)
        await account.get_vehicles()

        mock_api.get("/eadrax-vcs/v1/vehicles").respond(
            500, text=load_response(RESPONSE_DIR / "auth" / "auth_error_internal_error.txt")
        )
        with pytest.raises(httpx.HTTPStatusError):
            await account.get_vehicles()

    files = list(tmp_path.iterdir())
    json_files = [f for f in files if f.suffix == ".json"]
    txt_files = [f for f in files if f.suffix == ".txt"]

    assert len(json_files) == 2
    assert len(txt_files) == 2


@pytest.mark.asyncio
async def test_set_observer_value():
    """Test set_observer_position with valid arguments."""
    account = await get_mocked_account()

    account.set_observer_position(1.0, 2.0)

    assert account.observer_position == GPSPosition(1.0, 2.0)


@pytest.mark.asyncio
async def test_set_observer_not_set():
    """Test set_observer_position with no arguments."""
    account = await get_mocked_account()

    assert account.observer_position is None

    account.set_observer_position(17.99, 179.9)

    assert account.observer_position == GPSPosition(17.99, 179.9)


@pytest.mark.asyncio
async def test_set_observer_invalid_values():
    """Test set_observer_position with invalid arguments."""
    account = await get_mocked_account()

    with pytest.raises(TypeError):
        account.set_observer_position(None, 2.0)

    with pytest.raises(TypeError):
        account.set_observer_position(1.0, None)

    with pytest.raises(TypeError):
        account.set_observer_position(1.0, "16.0")


@account_mock()
@pytest.mark.asyncio
async def test_base_authentication():
    """Test logging in with the Authentication base clas."""
    account = get_account()
    account.mybmw_client_config = MyBMWClientConfiguration(Authentication(TEST_USERNAME, TEST_PASSWORD, TEST_REGION))
    with pytest.raises(NotImplementedError):
        await account.get_vehicles()


@account_mock()
@pytest.mark.asyncio
async def test_deprecated_account(caplog):
    """Test deprecation warning for ConnectedDriveAccount."""
    account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
    await account.get_vehicles()
    assert account is not None

    assert 1 == len(get_deprecation_warning_count(caplog))



"""Generic API management."""

import pathlib
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Optional

import httpx

from bimmer_connected.api.authentication import MyBMWAuthentication
from bimmer_connected.api.regions import get_server_url
from bimmer_connected.api.utils import log_to_to_file, raise_for_status_event_handler
from bimmer_connected.const import X_USER_AGENT, CarBrands


@dataclass
class MyBMWClientConfiguration:
    """Stores global settings for MyBMWClient."""

    authentication: MyBMWAuthentication
    log_response_path: Optional[pathlib.Path] = None


class MyBMWClient(httpx.AsyncClient):
    """Async HTTP client based on `httpx.AsyncClient` with automated OAuth token refresh."""

    def __init__(self, config: MyBMWClientConfiguration, *args, brand: CarBrands = None, **kwargs):
        self.config = config

        # Set default values
        kwargs["base_url"] = kwargs.get("base_url") or get_server_url(config.authentication.region)
        kwargs["headers"] = kwargs.get("headers") or self.generate_default_header(brand)

        # Register event hooks
        kwargs["event_hooks"] = defaultdict(list, **kwargs.get("event_hooks", {}))

        # Event hook which checks and updates token if required before request is sent
        async def update_request_header(request: httpx.Request):
            request.headers["authorization"] = await self.config.authentication.get_authentication()

        kwargs["event_hooks"]["request"].append(update_request_header)

        # Event hook for logging content to file
        async def log_response(response: httpx.Response):
            content = await response.aread()
            brand = [x for x in [b.value for b in CarBrands] if x in response.request.headers.get("x-user-agent", "")]
            base_file_name = "_".join([response.url.path.split("/")[-1]] + brand)
            log_to_to_file(content, config.log_response_path, base_file_name)  # type: ignore[arg-type]

        if config.log_response_path:
            kwargs["event_hooks"]["response"].append(log_response)

        # Event hook which calls raise_for_status on all requests
        kwargs["event_hooks"]["response"].append(raise_for_status_event_handler)

        super().__init__(*args, **kwargs)

    @staticmethod
    def generate_default_header(brand: CarBrands = None) -> Dict[str, str]:
        """Generate a header for HTTP requests to the server."""
        return {
            "accept": "application/json",
            "accept-language": "en",
            "user-agent": "Dart/2.13 (dart:io)",
            "x-user-agent": X_USER_AGENT.format(brand or CarBrands.BMW),
        }



"""Access to a MyBMW account and all vehicles therein."""

import asyncio
import datetime
import logging
import pathlib
from dataclasses import InitVar, dataclass, field
from typing import List, Optional

import httpx

from bimmer_connected.api.authentication import MyBMWAuthentication
from bimmer_connected.api.client import MyBMWClient, MyBMWClientConfiguration
from bimmer_connected.api.regions import Regions
from bimmer_connected.const import VEHICLES_URL, CarBrands
from bimmer_connected.utils import deprecated
from bimmer_connected.vehicle import MyBMWVehicle
from bimmer_connected.vehicle.models import GPSPosition

VALID_UNTIL_OFFSET = datetime.timedelta(seconds=10)

_LOGGER = logging.getLogger(__name__)
# lock = asyncio.Lock()


@dataclass
class MyBMWAccount:  # pylint: disable=too-many-instance-attributes
    """Create a new connection to the MyBMW web service."""

    username: str
    """MyBMW user name (email) or 86-prefixed phone number (China only)."""

    password: InitVar[str]
    """MyBMW password."""

    region: Regions
    """Region of the account. See `api.Regions`."""

    mybmw_client_config: MyBMWClientConfiguration = None  # type: ignore[assignment]
    """Optional. If provided, username/password/region are ignored."""

    log_responses: InitVar[pathlib.Path] = None
    """Optional. If set, all responses from the server will be logged to this directory."""

    observer_position: Optional[GPSPosition] = None
    """Optional. Required for getting a position on older cars."""

    vehicles: List[MyBMWVehicle] = field(default_factory=list, init=False)

    def __post_init__(self, password, log_responses):
        if self.mybmw_client_config is None:
            self.mybmw_client_config = MyBMWClientConfiguration(
                MyBMWAuthentication(self.username, password, self.region),
                log_response_path=log_responses,
            )

    async def get_vehicles(self) -> None:
        """Retrieve vehicle data from BMW servers."""
        _LOGGER.debug("Getting vehicle list")

        async with MyBMWClient(self.mybmw_client_config) as client:
            vehicles_request_params = {
                "apptimezone": self.utcdiff,
                "appDateTime": int(datetime.datetime.now().timestamp() * 1000),
                "tireGuardMode": "ENABLED",
            }
            vehicles_tasks: List[asyncio.Task] = []
            for brand in CarBrands:
                vehicles_tasks.append(
                    asyncio.ensure_future(
                        client.get(
                            VEHICLES_URL,
                            params=vehicles_request_params,
                            headers=client.generate_default_header(brand),
                        )
                    )
                )
            vehicles_responses: List[httpx.Response] = await asyncio.gather(*vehicles_tasks)

            for response in vehicles_responses:
                for vehicle_dict in response.json():
                    # If vehicle already exists, just update it's state
                    existing_vehicle = self.get_vehicle(vehicle_dict["vin"])
                    if existing_vehicle:
                        existing_vehicle.update_state(vehicle_dict)
                    else:
                        self.vehicles.append(MyBMWVehicle(self, vehicle_dict))

    def get_vehicle(self, vin: str) -> Optional[MyBMWVehicle]:
        """Get vehicle with given VIN.

        The search is NOT case sensitive.
        :param vin: VIN of the vehicle you want to get.
        :return: Returns None if no vehicle is found.
        """
        for car in self.vehicles:
            if car.vin.upper() == vin.upper():
                return car
        return None

    def set_observer_position(self, latitude: float, longitude: float) -> None:
        """Set the position of the observer for all vehicles."""
        self.observer_position = GPSPosition(latitude=latitude, longitude=longitude)

    @property
    def timezone(self):
        """Returns the current tzinfo."""
        return datetime.datetime.now().astimezone().tzinfo

    @property
    def utcdiff(self):
        """Returns the difference to UTC in minutes."""
        return round(self.timezone.utcoffset(datetime.datetime.now()).seconds / 60, 0)


@deprecated("MyBMWAccount")
class ConnectedDriveAccount(MyBMWAccount):
    """Deprecated class name for compatibility."""



"""Generals models used for bimmer_connected."""

import datetime
import logging
from dataclasses import dataclass
from typing import Any, Dict, Optional

from bimmer_connected.const import Regions
from bimmer_connected.coord_convert import gcj2wgs
from bimmer_connected.utils import parse_datetime
from bimmer_connected.vehicle.models import GPSPosition, VehicleDataBase

_LOGGER = logging.getLogger(__name__)


@dataclass
class VehicleLocation(VehicleDataBase):
    """The current position of a vehicle."""

    location: GPSPosition = GPSPosition(None, None)
    """The last known position of the vehicle."""

    heading: Optional[int] = None
    """The last known heading/direction of the vehicle."""

    vehicle_update_timestamp: Optional[datetime.datetime] = None
    account_region: Optional[Regions] = None
    remote_service_position: Optional[Dict] = None

    # pylint:disable=arguments-differ
    @classmethod
    def from_vehicle_data(cls, vehicle_data: Dict):
        """Creates the class based on vehicle data from API."""
        parsed = cls._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return cls(**parsed)
        return None

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict):
        date_dummy = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)

        retval: Dict[str, Any] = {}
        retval["vehicle_update_timestamp"] = max(
            parse_datetime(vehicle_data.get("properties", {}).get("lastUpdatedAt")) or date_dummy,
            parse_datetime(vehicle_data.get("status", {}).get("lastUpdatedAt")) or date_dummy,
        )
        if "properties" in vehicle_data and "vehicleLocation" in vehicle_data["properties"]:
            location = vehicle_data["properties"]["vehicleLocation"]
            retval["location"] = GPSPosition(location["coordinates"]["latitude"], location["coordinates"]["longitude"])
            retval["heading"] = location["heading"]
        return retval

    def _update_after_parse(self, parsed: Dict) -> Dict:
        """Updates parsed vehicle data with attributes stored in class if needed."""
        retval = parsed
        # Overwrite vehicle data with remote service position if available & newer
        if self.remote_service_position is not None:
            t_remote = self.remote_service_position.get(
                "timestamp", datetime.datetime(1900, 1, 1, tzinfo=datetime.timezone.utc)
            )
            if t_remote > self.vehicle_update_timestamp:
                retval["location"] = GPSPosition(
                    self.remote_service_position["latitude"], self.remote_service_position["longitude"]
                )
                retval["heading"] = self.remote_service_position["heading"]

        # Convert GCJ02 to WGS84 for positions in China
        if self.account_region == Regions.CHINA and "location" in retval and retval["location"].latitude is not None:
            gcj_lon, gcj_lat = gcj2wgs(gcjLon=retval["location"].longitude, gcjLat=retval["location"].latitude)
            retval["location"] = GPSPosition(gcj_lat, gcj_lon)
        return retval

    def set_remote_service_position(self, remote_service_dict: Dict):
        """Store remote service position returned from vehicle finder service."""
        if remote_service_dict.get("errorDetails"):
            error = remote_service_dict["errorDetails"]
            _LOGGER.error("Error retrieving vehicle position. %s: %s", error["title"], error["description"])
        else:
            pos = remote_service_dict["positionData"]["position"]
            pos["timestamp"] = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)

            self.remote_service_position = pos

            self.__dict__.update(self._update_after_parse({}))



"""Models the charging profiles of a vehicle."""

import datetime
import logging
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from bimmer_connected.vehicle.models import StrEnum, VehicleDataBase

_LOGGER = logging.getLogger(__name__)


class ChargingMode(StrEnum):
    """Charging mode of electric vehicle."""

    IMMEDIATE_CHARGING = "immediateCharging"
    DELAYED_CHARGING = "delayedCharging"


class ChargingPreferences(StrEnum):
    """Charging preferences of electric vehicle."""

    NO_PRESELECTION = "noPreSelection"
    CHARGING_WINDOW = "chargingWindow"


class TimerTypes(StrEnum):
    """Different timer types."""

    TWO_WEEKS = "twoWeeksTimer"
    ONE_WEEK = "weeklyPlanner"
    OVERRIDE_TIMER = "overrideTimer"


class ChargingWindow:
    """A charging window."""

    def __init__(self, window_dict: dict):
        self._window_dict = window_dict

    @property
    def start_time(self) -> datetime.time:
        """Start of the charging window."""
        # end of reductionOfChargeCurrent == start of charging window
        return datetime.time(int(self._window_dict["end"]["hour"]), int(self._window_dict["end"]["minute"]))

    @property
    def end_time(self) -> datetime.time:
        """End of the charging window."""
        # start of reductionOfChargeCurrent == end of charging window
        return datetime.time(int(self._window_dict["start"]["hour"]), int(self._window_dict["start"]["minute"]))


class DepartureTimer:
    """A departure timer."""

    def __init__(self, timer_dict: dict):
        self._timer_dict: Dict = timer_dict

    @property
    def timer_id(self) -> Optional[int]:
        """ID of this timer."""
        return self._timer_dict.get("id")

    @property
    def start_time(self) -> Optional[datetime.time]:
        """Deperture time for this timer."""
        if "timeStamp" not in self._timer_dict:
            return None
        return datetime.time(int(self._timer_dict["timeStamp"]["hour"]), int(self._timer_dict["timeStamp"]["minute"]))

    @property
    def action(self) -> Optional[str]:
        """What does the timer do."""
        return self._timer_dict.get("action")

    @property
    def weekdays(self) -> List[str]:
        """Active weekdays for this timer."""
        return self._timer_dict.get("timerWeekDays")  # type: ignore


@dataclass
class ChargingProfile(VehicleDataBase):  # pylint:disable=too-many-instance-attributes
    """Models the charging profile of a vehicle."""

    is_pre_entry_climatization_enabled: bool
    """Get status of pre-entry climatization."""

    timer_type: TimerTypes
    """Returns the current timer plan type."""

    departure_times: List[DepartureTimer]
    """List of timers."""

    preferred_charging_window: ChargingWindow
    """Returns the preferred charging window."""

    charging_preferences: ChargingPreferences
    """Returns the preferred charging preferences."""

    charging_mode: ChargingMode
    """Returns the preferred charging mode."""

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Dict:
        """Parse doors and windows."""
        retval: Dict[str, Any] = {}

        if "status" not in vehicle_data or "chargingProfile" not in vehicle_data["status"]:
            if vehicle_data["capabilities"]["isChargingPlanSupported"]:
                _LOGGER.error("Unable to read data from `status.chargingProfile`.")
            return retval

        charging_profile = vehicle_data["status"]["chargingProfile"]

        retval["is_pre_entry_climatization_enabled"] = bool(charging_profile["climatisationOn"])
        retval["departure_times"] = [DepartureTimer(t) for t in charging_profile["departureTimes"]]
        retval["preferred_charging_window"] = ChargingWindow(charging_profile["reductionOfChargeCurrent"])
        retval["timer_type"] = TimerTypes(charging_profile["chargingControlType"])
        retval["charging_preferences"] = ChargingPreferences(charging_profile["chargingPreference"])
        retval["charging_mode"] = ChargingMode(charging_profile["chargingMode"])

        return retval



"""Models state and remote services of one vehicle."""
import datetime
import logging
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Type

from bimmer_connected.api.client import MyBMWClient
from bimmer_connected.const import SERVICE_PROPERTIES, SERVICE_STATUS, VEHICLE_IMAGE_URL, CarBrands
from bimmer_connected.utils import deprecated, parse_datetime
from bimmer_connected.vehicle.charging_profile import ChargingProfile
from bimmer_connected.vehicle.doors_windows import DoorsAndWindows
from bimmer_connected.vehicle.fuel_and_battery import FuelAndBattery
from bimmer_connected.vehicle.location import VehicleLocation
from bimmer_connected.vehicle.models import StrEnum, ValueWithUnit
from bimmer_connected.vehicle.remote_services import RemoteServices
from bimmer_connected.vehicle.reports import CheckControlMessageReport, ConditionBasedServiceReport
from bimmer_connected.vehicle.vehicle_status import VehicleStatus

if TYPE_CHECKING:
    from bimmer_connected.account import MyBMWAccount
    from bimmer_connected.vehicle.models import VehicleDataBase


_LOGGER = logging.getLogger(__name__)


class DriveTrainType(StrEnum):
    """Different types of drive trains."""

    COMBUSTION = "COMBUSTION"
    PLUGIN_HYBRID = "PLUGIN_HYBRID"  # PHEV
    ELECTRIC = "ELECTRIC"
    HYBRID = "HYBRID"  # mild hybrids


#: Set of drive trains that have a combustion engine
COMBUSTION_ENGINE_DRIVE_TRAINS = {DriveTrainType.COMBUSTION, DriveTrainType.PLUGIN_HYBRID}

#: set of drive trains that have a high voltage battery
HV_BATTERY_DRIVE_TRAINS = {DriveTrainType.PLUGIN_HYBRID, DriveTrainType.ELECTRIC}


class VehicleViewDirection(StrEnum):
    """Viewing angles for the vehicle.

    This is used to get a rendered image of the vehicle.
    """

    FRONTSIDE = "VehicleStatus"
    FRONT = "VehicleInfo"
    # REARSIDE = 'REARSIDE'
    # REAR = 'REAR'
    SIDE = "ChargingHistory"
    # DASHBOARD = 'DASHBOARD'
    # DRIVERDOOR = 'DRIVERDOOR'
    # REARBIRDSEYE = 'REARBIRDSEYE'


class LscType(StrEnum):
    """Known Values for lsc_type field.

    Not really sure, what this value really contains.
    """

    NOT_CAPABLE = "NOT_CAPABLE"
    NOT_SUPPORTED = "NOT_SUPPORTED"
    ACTIVATED = "ACTIVATED"


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class MyBMWVehicle:
    """Models state and remote services of one vehicle.

    :param account: MyBMW account this vehicle belongs to
    :param attributes: attributes of the vehicle as provided by the server
    """

    def __init__(self, account: "MyBMWAccount", vehicle_data: dict) -> None:
        """Initializes a MyBMWVehicle."""
        self.account = account
        self.data = vehicle_data
        self.status = VehicleStatus(self)
        self.remote_services = RemoteServices(self)
        self.fuel_and_battery: FuelAndBattery = FuelAndBattery(account_timezone=account.timezone)
        self.vehicle_location: VehicleLocation = VehicleLocation(account_region=account.region)
        self.doors_and_windows: DoorsAndWindows = DoorsAndWindows()
        self.condition_based_services: ConditionBasedServiceReport = ConditionBasedServiceReport()
        self.check_control_messages: CheckControlMessageReport = CheckControlMessageReport()
        self.charging_profile: Optional[ChargingProfile] = None

        self.update_state(vehicle_data)

    def update_state(self, vehicle_data) -> None:
        """Update the state of a vehicle."""
        self.data = vehicle_data

        update_entities: List[Tuple[Type["VehicleDataBase"], str]] = [
            (FuelAndBattery, "fuel_and_battery"),
            (VehicleLocation, "vehicle_location"),
            (DoorsAndWindows, "doors_and_windows"),
            (ConditionBasedServiceReport, "condition_based_services"),
            (CheckControlMessageReport, "check_control_messages"),
            (ChargingProfile, "charging_profile"),
        ]
        for cls, vehicle_attribute in update_entities:
            if getattr(self, vehicle_attribute) is None:
                setattr(self, vehicle_attribute, cls.from_vehicle_data(vehicle_data))
            else:
                curr_attr: "VehicleDataBase" = getattr(self, vehicle_attribute)
                curr_attr.update_from_vehicle_data(vehicle_data)

    # # # # # # # # # # # # # # #
    # Generic attributes
    # # # # # # # # # # # # # # #

    @property
    def _status(self) -> Dict:
        """A shortcut to `data.status`."""
        return self.data[SERVICE_STATUS]

    @property
    def _properties(self) -> Dict:
        """A shortcut to `data.properties`."""
        return self.data[SERVICE_PROPERTIES]

    @property
    def brand(self) -> CarBrands:
        """Get the car brand."""
        return CarBrands(self.data["brand"])

    @property
    def name(self) -> str:
        """Get the name of the vehicle."""
        return self.data["model"]

    @property
    def vin(self) -> str:
        """Get the VIN (vehicle identification number) of the vehicle."""
        return self.data["vin"]

    @property
    def drive_train(self) -> DriveTrainType:
        """Get the type of drive train of the vehicle."""
        return DriveTrainType(self.data["driveTrain"])

    @property
    def mileage(self) -> ValueWithUnit:
        """Get the mileage of the vehicle."""
        return ValueWithUnit(self._status["currentMileage"]["mileage"], self._status["currentMileage"]["units"])

    @property
    def timestamp(self) -> Optional[datetime.datetime]:
        """Get the timestamp when the data was recorded."""
        timestamps = [
            ts
            for ts in [
                parse_datetime(self._properties.get("lastUpdatedAt")),  # type: ignore
                parse_datetime(self._status.get("lastUpdatedAt")),  # type: ignore
            ]
            if ts
        ]
        if len(timestamps) == 0:
            return None
        return max(timestamps)

    @property
    def last_update_reason(self) -> str:
        """The reason for the last state update."""
        return self._status["timestampMessage"]

    # # # # # # # # # # # # # # #
    # Capabilities & properties
    # # # # # # # # # # # # # # #

    @property
    def has_electric_drivetrain(self) -> bool:
        """Return True if vehicle is equipped with a high voltage battery.

        In this case we can get the state of the battery in the state attributes.
        """
        return self.drive_train in HV_BATTERY_DRIVE_TRAINS

    @property
    def has_range_extender_drivetrain(self) -> bool:
        """Return True if vehicle is equipped with a range extender.

        In this case we can get the state of the gas tank."""
        return self.drive_train == DriveTrainType.ELECTRIC and self.fuel_indicator_count == 3

    @property
    def has_combustion_drivetrain(self) -> bool:
        """Return True if vehicle is equipped with an internal combustion engine.

        In this case we can get the state of the gas tank."""
        return self.drive_train in COMBUSTION_ENGINE_DRIVE_TRAINS

    @property
    def is_charging_plan_supported(self) -> bool:
        """Return True if charging control (weekly planner) is available."""
        return self.data["capabilities"]["isChargingPlanSupported"]

    @property
    def is_vehicle_tracking_enabled(self) -> bool:
        """Return True if vehicle finder is enabled in vehicle."""
        return self.data["capabilities"]["vehicleFinder"]["isEnabled"]

    @property
    def is_vehicle_active(self) -> bool:
        """Check if the vehicle is active/moving.

        If the vehicle was active/moving at the time of the last status update, current position is not available.
        """
        return self._properties["inMotion"]

    @property
    def is_lsc_enabled(self) -> bool:
        """Return True if LastStateCall is enabled (vehicle automatically updates API)."""
        return self.data["capabilities"]["lastStateCall"]["lscState"] == "ACTIVATED"

    @property
    def drive_train_attributes(self) -> List[str]:
        """Get list of attributes available for the drive train of the vehicle.

        The list of available attributes depends if on the type of drive train.
        Some attributes only exist for electric/hybrid vehicles, others only if you
        have a combustion engine. Depending on the state of the vehicle, some of
        the attributes might still be None.
        """
        result = ["remaining_range_total", "mileage"]
        if self.has_electric_drivetrain:
            result += [
                "charging_time_remaining",
                "charging_start_time",
                "charging_end_time",
                "charging_time_label",
                "charging_status",
                "charging_level_hv",
                "connection_status",
                "remaining_range_electric",
                "last_charging_end_result",
            ]
        if self.has_combustion_drivetrain or self.has_range_extender_drivetrain:
            result += ["remaining_fuel", "remaining_range_fuel", "fuel_percent"]
        return result

    @property
    def lsc_type(self) -> LscType:
        """Get the lscType of the vehicle.

        Not really sure what that value really means. If it is NOT_CAPABLE, that probably means that the
        vehicle state will not contain much data.
        """
        return LscType(self.data["capabilities"]["lastStateCall"].get("lscState"))

    @property
    def available_attributes(self) -> List[str]:
        """Get the list of non-drivetrain attributes available for this vehicle."""
        # attributes available in all vehicles
        result = ["gps_position", "vin"]
        if self.lsc_type == LscType.ACTIVATED:
            # generic attributes if lsc_type =! NOT_SUPPORTED
            result += self.drive_train_attributes
            result += [
                "condition_based_services",
                "check_control_messages",
                "door_lock_state",
                "timestamp",
                "last_update_reason",
            ]
            # required for existing Home Assistant binary sensors
            result += ["lids", "windows"]
        return result

    @property
    def fuel_indicator_count(self) -> int:
        """Gets the number of fuel indicators.

        Can be used to identify REX vehicles if driveTrain == ELECTRIC.
        """
        return len(self._status["fuelIndicators"])

    # # # # # # # # # # # # # # #
    # Generic functions
    # # # # # # # # # # # # # # #

    async def get_vehicle_image(self, direction: VehicleViewDirection) -> bytes:
        """Get a rendered image of the vehicle.

        :returns bytes containing the image in PNG format.
        """
        url = VEHICLE_IMAGE_URL.format(
            vin=self.vin,
            view=direction.value,
        )
        # the accept field of the header needs to be updated as we want a png not the usual JSON
        async with MyBMWClient(self.account.mybmw_client_config, brand=self.brand) as client:
            response = await client.get(url, headers={"accept": "image/png"})
        return response.content

    # # # # # # # # # # # # # # #
    # Deprecated
    # # # # # # # # # # # # # # #

    @property
    @deprecated("vehicle.has_electric_drivetrain")
    def has_hv_battery(self) -> bool:
        # pylint:disable=missing-function-docstring
        return self.has_electric_drivetrain

    @property
    @deprecated("vehicle.has_range_extender_drivetrain")
    def has_range_extender(self) -> bool:
        # pylint:disable=missing-function-docstring
        return self.has_range_extender_drivetrain

    @property
    @deprecated("vehicle.has_combustion_drivetrain")
    def has_internal_combustion_engine(self) -> bool:
        # pylint:disable=missing-function-docstring
        return self.has_combustion_drivetrain

    @property
    @deprecated("vehicle.is_charging_plan_supported")
    def has_weekly_planner_service(self) -> bool:
        # pylint:disable=missing-function-docstring
        return self.is_charging_plan_supported


@deprecated("MyBMWVehicle")
class ConnectedDriveVehicle(MyBMWVehicle):
    """Deprecated class name for compatibility."""



"""Models the state of a vehicle."""

import datetime
import logging
from typing import TYPE_CHECKING, Dict, List, Optional

from bimmer_connected.utils import deprecated
from bimmer_connected.vehicle.models import GPSPosition, ValueWithUnit

if TYPE_CHECKING:
    from bimmer_connected.vehicle import MyBMWVehicle
    from bimmer_connected.vehicle.doors_windows import Lid, LockState, Window
    from bimmer_connected.vehicle.fuel_and_battery import ChargingState
    from bimmer_connected.vehicle.reports import CheckControlMessage, ConditionBasedService

_LOGGER = logging.getLogger(__name__)


class VehicleStatus:  # pylint: disable=too-many-public-methods
    """Models the status of a vehicle."""

    # pylint: disable=unused-argument
    def __init__(self, vehicle: "MyBMWVehicle", status_dict: Dict = None):
        """Constructor."""
        self.vehicle = vehicle

    @property
    @deprecated("vehicle.timestamp")
    def timestamp(self) -> Optional[datetime.datetime]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.timestamp

    @property
    @deprecated("vehicle.vehicle_location.location")
    def gps_position(self) -> GPSPosition:
        # pylint:disable=missing-function-docstring
        return self.vehicle.vehicle_location.location

    @property
    @deprecated("vehicle.vehicle_location.heading")
    def gps_heading(self) -> Optional[int]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.vehicle_location.heading

    @property
    @deprecated("vehicle.is_vehicle_active")
    def is_vehicle_active(self) -> bool:
        # pylint:disable=missing-function-docstring
        return self.vehicle.is_vehicle_active

    @property
    @deprecated("vehicle.mileage")
    def mileage(self) -> ValueWithUnit:
        # pylint:disable=missing-function-docstring
        return self.vehicle.mileage

    @property
    @deprecated("vehicle.fuel_and_battery.remaining_range_fuel")
    def remaining_range_fuel(self) -> Optional[ValueWithUnit]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.fuel_and_battery.remaining_range_fuel

    @property
    @deprecated("vehicle.fuel_and_battery.remaining_fuel")
    def remaining_fuel(self) -> Optional[ValueWithUnit]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.fuel_and_battery.remaining_fuel

    @property
    @deprecated("vehicle.fuel_indicator_count")
    def fuel_indicator_count(self) -> Optional[int]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.fuel_indicator_count

    @property
    @deprecated("vehicle.doors_and_windows.lids")
    def lids(self) -> List["Lid"]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.doors_and_windows.lids

    @property
    @deprecated("vehicle.doors_and_windows.open_lids")
    def open_lids(self) -> List["Lid"]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.doors_and_windows.open_lids

    @property
    @deprecated("vehicle.doors_and_windows.all_lids_closed")
    def all_lids_closed(self) -> bool:
        # pylint:disable=missing-function-docstring
        return self.vehicle.doors_and_windows.all_lids_closed

    @property
    @deprecated("vehicle.doors_and_windows.windows")
    def windows(self) -> List["Window"]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.doors_and_windows.windows

    @property
    @deprecated("vehicle.doors_and_windows.open_windows")
    def open_windows(self) -> List["Window"]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.doors_and_windows.open_windows

    @property
    @deprecated("vehicle.doors_and_windows.all_lids_closed")
    def all_windows_closed(self) -> bool:
        # pylint:disable=missing-function-docstring
        return self.vehicle.doors_and_windows.all_lids_closed

    @property
    @deprecated("vehicle.doors_and_windows.door_lock_state")
    def door_lock_state(self) -> "LockState":
        # pylint:disable=missing-function-docstring
        return self.vehicle.doors_and_windows.door_lock_state

    @property
    @deprecated("vehicle.last_update_reason")
    def last_update_reason(self) -> str:
        # pylint:disable=missing-function-docstring
        return self.vehicle.last_update_reason

    @property
    @deprecated()
    def last_charging_end_result(self) -> None:
        # pylint:disable=missing-function-docstring
        return None  # Not available in My BMW

    @property
    @deprecated("vehicle.fuel_and_battery.is_charger_connected")
    def connection_status(self) -> str:
        # pylint:disable=missing-function-docstring
        return "CONNECTED" if self.vehicle.fuel_and_battery.is_charger_connected else "DISCONNECTED"

    @property
    @deprecated("vehicle.condition_based_services.messages")
    def condition_based_services(self) -> List["ConditionBasedService"]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.condition_based_services.messages

    @property
    @deprecated("vehicle.condition_based_services.is_service_required")
    def are_all_cbs_ok(self) -> bool:
        # pylint:disable=missing-function-docstring
        return not self.vehicle.condition_based_services.is_service_required

    @property
    @deprecated()
    def parking_lights(self) -> None:
        # pylint:disable=missing-function-docstring
        return None  # Not available in My BMW

    @property
    @deprecated()
    def has_parking_light_state(self) -> bool:
        # pylint:disable=missing-function-docstring
        return False  # Not available in My BMW

    @property
    @deprecated()
    def are_parking_lights_on(self) -> None:
        # pylint:disable=missing-function-docstring
        return None  # Not available in My BMW

    @property
    @deprecated("vehicle.fuel_and_battery.remaining_range_electric")
    def remaining_range_electric(self) -> Optional[ValueWithUnit]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.fuel_and_battery.remaining_range_electric

    @property
    @deprecated("vehicle.fuel_and_battery.remaining_range_total")
    def remaining_range_total(self) -> Optional[ValueWithUnit]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.fuel_and_battery.remaining_range_total

    @property
    @deprecated()
    def max_range_electric(self) -> Optional[int]:
        # pylint:disable=missing-function-docstring
        return None  # Not available in My BMW

    @property
    @deprecated("vehicle.fuel_and_battery.charging_status")
    def charging_status(self) -> Optional["ChargingState"]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.fuel_and_battery.charging_status

    @property
    @deprecated()
    def charging_time_remaining(self) -> None:
        # pylint:disable=missing-function-docstring
        return None

    @property
    @deprecated("vehicle.fuel_and_battery.charging_start_time")
    def charging_start_time(self) -> Optional[datetime.datetime]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.fuel_and_battery.charging_start_time

    @property
    @deprecated("vehicle.fuel_and_battery.charging_end_time")
    def charging_end_time(self) -> Optional[datetime.datetime]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.fuel_and_battery.charging_end_time

    @property
    @deprecated("vehicle.fuel_and_battery.charging_time_label")
    def charging_time_label(self) -> Optional[str]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.fuel_and_battery.charging_time_label

    @property
    @deprecated("vehicle.fuel_and_battery.remaining_battery_percent")
    def charging_level_hv(self) -> Optional[int]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.fuel_and_battery.remaining_battery_percent

    @property
    @deprecated("vehicle.fuel_and_battery.remaining_fuel_percent")
    def fuel_percent(self) -> Optional[int]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.fuel_and_battery.remaining_fuel_percent

    @property
    @deprecated("vehicle.check_control_messages.messages")
    def check_control_messages(self) -> List["CheckControlMessage"]:
        # pylint:disable=missing-function-docstring
        return self.vehicle.check_control_messages.messages

    @property
    @deprecated("vehicle.check_control_messages.has_check_control_messages")
    def has_check_control_messages(self) -> bool:
        # pylint:disable=missing-function-docstring
        return self.vehicle.check_control_messages.has_check_control_messages



"""Models the state of a vehicle."""

import datetime
import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from bimmer_connected.utils import parse_datetime
from bimmer_connected.vehicle.models import StrEnum, ValueWithUnit, VehicleDataBase

_LOGGER = logging.getLogger(__name__)


class ConditionBasedServiceStatus(StrEnum):
    """Status of the condition based services."""

    OK = "OK"
    OVERDUE = "OVERDUE"
    PENDING = "PENDING"


@dataclass
class ConditionBasedService:  # pylint: disable=too-few-public-methods
    """Entry in the list of condition based services."""

    service_type: str
    state: ConditionBasedServiceStatus
    due_date: Optional[datetime.date]
    due_distance: ValueWithUnit

    # pylint:disable=invalid-name,unused-argument,redefined-builtin
    @classmethod
    def from_api_entry(cls, type: str, status: str, dateTime: str = None, distance: Dict = None, **kwargs):
        """Parse a condition based service entry from the API format to `ConditionBasedService`."""
        due_distance = ValueWithUnit(distance["value"], distance["units"]) if distance else ValueWithUnit(None, None)
        due_date = parse_datetime(dateTime) if dateTime else None
        return cls(type, ConditionBasedServiceStatus(status), due_date, due_distance)


@dataclass
class ConditionBasedServiceReport(VehicleDataBase):
    """Parses and summarizes condition based services (e.g. next oil service)."""

    messages: List[ConditionBasedService] = field(default_factory=list)
    """List of the condition based services."""

    is_service_required: bool = False
    """Indicates if a service is required."""

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Optional[Dict]:
        """Parse doors and windows."""
        retval: Dict[str, Any] = {}

        if "properties" not in vehicle_data or "serviceRequired" not in vehicle_data["properties"]:
            _LOGGER.error("Unable to read data from `properties.serviceRequired`.")
            return retval

        messages = vehicle_data["properties"]["serviceRequired"]
        retval["messages"] = [ConditionBasedService.from_api_entry(**m) for m in messages]
        retval["is_service_required"] = vehicle_data["properties"]["isServiceRequired"]

        return retval


class CheckControlStatus(StrEnum):
    """Status of the condition based services."""

    OK = "OK"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"


@dataclass
class CheckControlMessage:
    """Check control message sent from the server."""

    description_short: str
    description_long: str
    state: CheckControlStatus

    # pylint:disable=invalid-name,unused-argument
    @classmethod
    def from_api_entry(cls, title: str, longDescription: str, state: str, **kwargs):
        """Parses a check control entry from the API format to `CheckControlMessage`."""
        return cls(title, longDescription, CheckControlStatus(state))


@dataclass
class CheckControlMessageReport(VehicleDataBase):
    """Parses and summarizes check control messages (e.g. low tire pressure)."""

    messages: List[CheckControlMessage] = field(default_factory=list)
    """List of check control messages."""

    has_check_control_messages: bool = False
    """Indicates if check control messages are present."""

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Optional[Dict]:
        """Parse doors and windows."""
        retval: Dict[str, Any] = {}

        if "status" not in vehicle_data or "checkControlMessages" not in vehicle_data["status"]:
            _LOGGER.error("Unable to read data from `status.checkControlMessages`.")
            return retval

        messages = vehicle_data["status"]["checkControlMessages"]
        retval["messages"] = [CheckControlMessage.from_api_entry(**m) for m in messages if m["state"] != "OK"]
        retval["has_check_control_messages"] = vehicle_data["status"]["checkControlMessagesGeneralState"] != "No Issues"

        return retval
